# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'timestamp_dialog_alt.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_manualTimestampDialog(object):
    def setupUi(self, manualTimestampDialog):
        manualTimestampDialog.setObjectName("manualTimestampDialog")
        manualTimestampDialog.resize(863, 544)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(manualTimestampDialog.sizePolicy().hasHeightForWidth())
        manualTimestampDialog.setSizePolicy(sizePolicy)
        manualTimestampDialog.setSizeGripEnabled(True)
        self.gridLayout = QtWidgets.QGridLayout(manualTimestampDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.label_13 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_13.setObjectName("label_13")
        self.horizontalLayout_6.addWidget(self.label_13)
        self.gridLayout.addLayout(self.horizontalLayout_6, 0, 0, 1, 1)
        self.verticalLayout_10 = QtWidgets.QVBoxLayout()
        self.verticalLayout_10.setObjectName("verticalLayout_10")
        self.label_14 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_14.setObjectName("label_14")
        self.verticalLayout_10.addWidget(self.label_14)
        self.radioButtonNTSC = QtWidgets.QRadioButton(manualTimestampDialog)
        self.radioButtonNTSC.setChecked(True)
        self.radioButtonNTSC.setObjectName("radioButtonNTSC")
        self.verticalLayout_10.addWidget(self.radioButtonNTSC)
        self.radioButtonPAL = QtWidgets.QRadioButton(manualTimestampDialog)
        self.radioButtonPAL.setObjectName("radioButtonPAL")
        self.verticalLayout_10.addWidget(self.radioButtonPAL)
        self.line_2 = QtWidgets.QFrame(manualTimestampDialog)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.verticalLayout_10.addWidget(self.line_2)
        self.radioButtonCustom = QtWidgets.QRadioButton(manualTimestampDialog)
        self.radioButtonCustom.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.radioButtonCustom.setObjectName("radioButtonCustom")
        self.verticalLayout_10.addWidget(self.radioButtonCustom)
        self.gridLayout.addLayout(self.verticalLayout_10, 3, 0, 1, 1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_9 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_9.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft)
        self.label_9.setObjectName("label_9")
        self.horizontalLayout_3.addWidget(self.label_9)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(manualTimestampDialog)
        self.label.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.frameNum1 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.frameNum1.setObjectName("frameNum1")
        self.verticalLayout.addWidget(self.frameNum1)
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label_2 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_2.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_2.addWidget(self.label_2)
        self.hh1 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.hh1.setObjectName("hh1")
        self.verticalLayout_2.addWidget(self.hh1)
        self.horizontalLayout.addLayout(self.verticalLayout_2)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_3 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_3.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_3.addWidget(self.label_3)
        self.mm1 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.mm1.setObjectName("mm1")
        self.verticalLayout_3.addWidget(self.mm1)
        self.horizontalLayout.addLayout(self.verticalLayout_3)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setSpacing(0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label_4 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_4.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_4.addWidget(self.label_4)
        self.ss1 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.ss1.setObjectName("ss1")
        self.verticalLayout_4.addWidget(self.ss1)
        self.horizontalLayout.addLayout(self.verticalLayout_4)
        self.horizontalLayout_3.addLayout(self.horizontalLayout)
        self.gridLayout.addLayout(self.horizontalLayout_3, 1, 0, 1, 1)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_10 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_10.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft)
        self.label_10.setObjectName("label_10")
        self.horizontalLayout_4.addWidget(self.label_10)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout()
        self.verticalLayout_5.setSpacing(0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.label_5 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_5.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft)
        self.label_5.setObjectName("label_5")
        self.verticalLayout_5.addWidget(self.label_5)
        self.frameNum2 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.frameNum2.setObjectName("frameNum2")
        self.verticalLayout_5.addWidget(self.frameNum2)
        self.horizontalLayout_2.addLayout(self.verticalLayout_5)
        self.verticalLayout_6 = QtWidgets.QVBoxLayout()
        self.verticalLayout_6.setSpacing(0)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.label_6 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_6.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_6.setObjectName("label_6")
        self.verticalLayout_6.addWidget(self.label_6)
        self.hh2 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.hh2.setObjectName("hh2")
        self.verticalLayout_6.addWidget(self.hh2)
        self.horizontalLayout_2.addLayout(self.verticalLayout_6)
        self.verticalLayout_7 = QtWidgets.QVBoxLayout()
        self.verticalLayout_7.setSpacing(0)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.label_7 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_7.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_7.setObjectName("label_7")
        self.verticalLayout_7.addWidget(self.label_7)
        self.mm2 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.mm2.setObjectName("mm2")
        self.verticalLayout_7.addWidget(self.mm2)
        self.horizontalLayout_2.addLayout(self.verticalLayout_7)
        self.verticalLayout_8 = QtWidgets.QVBoxLayout()
        self.verticalLayout_8.setSpacing(0)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.label_8 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_8.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_8.setObjectName("label_8")
        self.verticalLayout_8.addWidget(self.label_8)
        self.ss2 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.ss2.setObjectName("ss2")
        self.verticalLayout_8.addWidget(self.ss2)
        self.horizontalLayout_2.addLayout(self.verticalLayout_8)
        self.horizontalLayout_4.addLayout(self.horizontalLayout_2)
        self.gridLayout.addLayout(self.horizontalLayout_4, 2, 0, 1, 1)
        self.buttonBox = QtWidgets.QDialogButtonBox(manualTimestampDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 5, 0, 1, 1)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_11 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_11.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.label_11.setObjectName("label_11")
        self.horizontalLayout_5.addWidget(self.label_11)
        self.verticalLayout_9 = QtWidgets.QVBoxLayout()
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.frameDeltaTime = QtWidgets.QLineEdit(manualTimestampDialog)
        self.frameDeltaTime.setCursorMoveStyle(QtCore.Qt.VisualMoveStyle)
        self.frameDeltaTime.setObjectName("frameDeltaTime")
        self.verticalLayout_9.addWidget(self.frameDeltaTime)
        self.horizontalLayout_5.addLayout(self.verticalLayout_9)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem)
        self.gridLayout.addLayout(self.horizontalLayout_5, 4, 0, 1, 1)

        self.retranslateUi(manualTimestampDialog)
        self.buttonBox.accepted.connect(manualTimestampDialog.accept)
        self.buttonBox.rejected.connect(manualTimestampDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(manualTimestampDialog)
        manualTimestampDialog.setTabOrder(self.frameNum1, self.hh1)
        manualTimestampDialog.setTabOrder(self.hh1, self.mm1)
        manualTimestampDialog.setTabOrder(self.mm1, self.ss1)
        manualTimestampDialog.setTabOrder(self.ss1, self.frameNum2)
        manualTimestampDialog.setTabOrder(self.frameNum2, self.hh2)
        manualTimestampDialog.setTabOrder(self.hh2, self.mm2)
        manualTimestampDialog.setTabOrder(self.mm2, self.ss2)
        manualTimestampDialog.setTabOrder(self.ss2, self.frameDeltaTime)

    def retranslateUi(self, manualTimestampDialog):
        _translate = QtCore.QCoreApplication.translate
        manualTimestampDialog.setWindowTitle(_translate("manualTimestampDialog", "Manual Timestamp Entry"))
        self.label_13.setText(_translate("manualTimestampDialog", "<html><head/><body><p>PyOTE assumes that a frame timestamp represents the UTC time of the synch pulse that <br/>first appears in the composite video output as the frame/field is emitted from the camera. <br/><br/>For an IOTA-VTI frame display, this will be the earliest of the two times displayed in the overlay text.</p><p>Note: if the light curve was processed in field mode, frame numbers are still used but become fractional<br/>to indicate fields.  And the frame time remains the proper time unit to specify.<br/></p></body></html>"))
        self.label_14.setText(_translate("manualTimestampDialog", "<html><head/><body><p><br/>In order to detect dropped readings, PyOTE needs to know the expected frame time. Make a selection below:<br/></p></body></html>"))
        self.radioButtonNTSC.setText(_translate("manualTimestampDialog", "use NTSC frame time (1.001/30 seconds)"))
        self.radioButtonPAL.setText(_translate("manualTimestampDialog", "use PAL    frame time (1.000/25 seconds)"))
        self.radioButtonCustom.setText(_translate("manualTimestampDialog", "use Custom frame time (enter number or numerical expression in box below)"))
        self.label_9.setText(_translate("manualTimestampDialog", "<html><head/><body><p>Timestamp (early frame)</p></body></html>"))
        self.label.setText(_translate("manualTimestampDialog", "frame number"))
        self.label_2.setText(_translate("manualTimestampDialog", "hh"))
        self.label_3.setText(_translate("manualTimestampDialog", "mm"))
        self.label_4.setText(_translate("manualTimestampDialog", "ss.ssss"))
        self.label_10.setText(_translate("manualTimestampDialog", "<html><head/><body><p>Timestamp (later frame)</p></body></html>"))
        self.label_5.setText(_translate("manualTimestampDialog", "frame number"))
        self.label_6.setText(_translate("manualTimestampDialog", "hh"))
        self.label_7.setText(_translate("manualTimestampDialog", "mm"))
        self.label_8.setText(_translate("manualTimestampDialog", "ss.ssss"))
        self.label_11.setText(_translate("manualTimestampDialog", "<html><head/><body><p>Custom frame time (seconds)</p></body></html>"))




from typing import Dict, Tuple
import matplotlib

matplotlib.use('Qt5Agg')
import matplotlib.pyplot as plt
from pyoteapp.genDiffraction import generalizedDiffraction, decide_model_to_use
import numpy as np


def fresnel_length_km(distance_AU: float, wavelength_nm: float = 500.0) -> float:
    """
    Calculates the fresnel length given the wavelength of light and distance to the object.

    :param distance_AU: distance to object in AU (Astronomical Units)
    :param wavelength_nm: wavelength of light in nanometers
    :return: fresnel length in km
    """

    # Convert distance_AU from Astronomical units to km
    distance_km = distance_AU * 149.6e6
    # Convert wavelength_nm to wavelength_km
    wavelength_km = wavelength_nm * 1e-12
    return np.sqrt(distance_km * wavelength_km / 2.0)


def generate_transition_point_time_correction_look_up_tables(
        baseline_intensity: float,
        event_intensity: float,
        frame_time_sec: float,
        asteroid_distance_AU: float = None,
        asteroid_diameter_km: float = None,
        shadow_speed_km_per_sec: float = None,
        centerline_offset=None,
        star_diameter_mas: float = None,
        d_limb_angle_degrees: float = 90.0,
        r_limb_angle_degrees: float = 90.0,
        skip_central_flash: bool = False
) -> Dict[str, np.ndarray]:
    """
    Compute D and R lookup tables for finding the time offset from the geometrical
    shadow position of an occultation given a valid transition point

    :param baseline_intensity: B (standard occultation terminology for intensity before occultation)
    :param event_intensity: A (standard occultation terminology for intensity during occultation)
    :param frame_time_sec: light accumulation (integration setting) of camera (seconds)
    :param asteroid_distance_AU: distance to the asteroid/occulting body (in Astronomical Units)
    :param shadow_speed_km_per_sec: speed of the shadow projected at the observing site (km / second)
    :param asteroid_diameter_km: asteroid diameter in km
    :param centerline_offset: distance of observation point from centerline of asteroid path (km)
    :param star_diameter_mas: diameter of star disk (mas - milliarcseconds)
    :param d_limb_angle_degrees: limb angle at disappearance edge (degrees - 90 degrees is head-on)
    :param r_limb_angle_degrees: limb angle at re-appearance edge (degrees - 90 degrees is head-on)
    :param skip_central_flash: set this to True if the (time consuming) cetral flash calculation is not needed
    :return:
    """

    RESOLUTION = 0.0001  # time resolution of lookup tables - 0.1 millisecond

    star_chords_r = None
    star_chords_d = None

    d_graze_values = None
    r_graze_values = None

    if asteroid_distance_AU is None:
        # We cannot calculate a diffraction model lightcurve due to insufficient information,
        # so we revert to an underlying square wave model lightcurve.  In addition, we cannot
        # take into account a finite star disk

        assert star_diameter_mas is None, \
            'Inconsistency: A star diameter was given without the required asteroid distance'

        time_range_seconds = np.ceil(frame_time_sec) + 1.0

        n_points = int(time_range_seconds / RESOLUTION)
        time_values = np.linspace(-time_range_seconds, time_range_seconds, 2 * n_points + 1)
        d_values = np.ndarray(time_values.size)
        r_values = np.ndarray(time_values.size)
        for i, value in enumerate(time_values):
            if value < 0.0:
                r_values[i] = event_intensity
                d_values[i] = baseline_intensity
            else:
                r_values[i] = baseline_intensity
                d_values[i] = event_intensity

        raw_d_values = d_values[:]
        raw_r_values = r_values[:]

        # Prepare the 'sample' that performs a box-car integration when convolved with the model lightcurve

        n_sample_points = round(frame_time_sec / RESOLUTION)
        sample = np.repeat(1.0 / n_sample_points, n_sample_points)

        # Convolve sample against lightcurve to compute the effect of camera frame-time integration.
        d_values = lightcurve_convolve(sample=sample, lightcurve=d_values, shift_needed=len(sample) - 1)
        r_values = lightcurve_convolve(sample=sample, lightcurve=r_values, shift_needed=len(sample) - 1)

        return {'time deltas': time_values, 'D curve': d_values, 'R curve': r_values,
                'star_chords_d': star_chords_d, 'star_chords_r': star_chords_r,
                'raw D': raw_d_values, 'raw R': raw_r_values, 'graze D': d_graze_values, 'graze R': r_graze_values,
                'star D': None, 'star R': None,
                'B': baseline_intensity, 'A': event_intensity}

    elif shadow_speed_km_per_sec is not None:

        # Here we introduce the new diffraction calculation
        x_avg, y_avg, rho, rho_wavelength, _, _, title = \
            generalizedDiffraction(asteroid_diam_km=asteroid_diameter_km,
                                   asteroid_distance_AU=asteroid_distance_AU,
                                   graze_offset_km=centerline_offset,
                                   wavelength1=400, wavelength2=600,
                                   skip_central_flash=skip_central_flash)

        raw_y_avg = np.copy(y_avg)

        y_avg = y_avg * baseline_intensity
        y_avg += event_intensity

        scaled_y_avg = np.copy(y_avg)

        time_values = x_avg / shadow_speed_km_per_sec

        # Prepare the 'sample' that performs a box-car integration when convolved with the model lightcurve
        # to produce the effect of the camera integration.
        n_sample_points = round(frame_time_sec / ((x_avg[1] - x_avg[0]) / shadow_speed_km_per_sec))
        n_sample_points = max(n_sample_points, 1)

        sample = np.repeat(1.0 / n_sample_points, n_sample_points)

        if star_diameter_mas is not None:
            # We are able to compose star chords to convolve with the curves found so far.
            # We have to do that separately for each limb because the limb angle could be different.
            star_chords_d, \
                star_chords_r, \
                star_chords_d_standalone, \
                star_chords_r_standalone = \
                get_star_chord_samples(star_diameter_mas=star_diameter_mas,
                                       distance_to_asteroid_AU=asteroid_distance_AU,
                                       shadow_speed=shadow_speed_km_per_sec,
                                       time_resolution=(x_avg[1] - x_avg[0]) / shadow_speed_km_per_sec,
                                       d_limb_angle_degrees=d_limb_angle_degrees,
                                       r_limb_angle_degrees=r_limb_angle_degrees,
                                       num_points_in_chord_base=len(y_avg)
                                       )
            # print(f'star_chords_d: {star_chords_d}')
            # Convolve sample against lightcurve to compute the effect of star chord integration.
            mid_point = len(y_avg) // 2
            d_values = y_avg[0:mid_point]
            d_values = lightcurve_convolve(sample=star_chords_d_standalone, lightcurve=d_values,
                                           shift_needed=len(star_chords_d_standalone) // 2)
            r_values = y_avg[mid_point: -1]
            r_values = lightcurve_convolve(sample=star_chords_r_standalone, lightcurve=r_values,
                                           shift_needed=len(star_chords_r_standalone) // 2)
            for i, value in enumerate(d_values):
                y_avg[i] = value

            for i, value in enumerate(r_values):
                y_avg[mid_point + i] = value

        # Convolve sample against lightcurve to compute the effect of camera frame-time integration.
        y_avg = lightcurve_convolve(sample=sample, lightcurve=y_avg, shift_needed=len(sample) - 1)

        return {'time deltas': time_values, 'camera response': y_avg,
                'star_chords_d': star_chords_d, 'star_chords_r': star_chords_r,
                'raw underlying': raw_y_avg, 'scaled underlying': scaled_y_avg,
                'B': baseline_intensity, 'A': event_intensity}


def intensity_at_time(data, time, edge_type):
    assert edge_type in ['D', 'R']

    if edge_type == 'D':
        if time <= data['time deltas'][0]:
            return data['D curve'][0]
        if time >= data['time deltas'][-1]:
            return data['D curve'][-1]
        for i, t in enumerate(data['time deltas']):
            if t >= time:
                return data['D curve'][i]
        return None  # This should never happen

    else:
        if time < data['time deltas'][0]:
            return data['R curve'][0]
        if time > data['time deltas'][-1]:
            return data['R curve'][-1]
        for i, t in enumerate(data['time deltas']):
            if t >= time:
                return data['R curve'][i]
        return None  # This should never happen


def lightcurve_convolve(sample: np.ndarray, lightcurve: np.ndarray, shift_needed: int) -> np.ndarray:
    """
    Computes the convolution of sample[] against lightcurve[] with corrections for leading edge
    effects and for the right shift of the normal convolution calculation.   We do a
    counter-shift so as to maintain our start-of-exposure convention.

    :param sample: usually contains either a 'box-car' (for frame integration) or a set of star chords
    :param lightcurve: the underlying lightcurve (a square wave or a diffraction curve)
    :param shift_needed: size of left shift desired
    :return: convolution of sample against lightcurve (with length = len(lightcurve)
    """

    # To 'eliminate' the leading edge effect inherent in np.convolve(), we insert a compensating number of
    # duplicate points at the beginning.  This works well only if the lightcurve has already achieved
    # 'steady state' in this region.  We strive to guarantee this everywhere this function is called.

    leading_edge_extension = np.ones(len(sample) - 1) * lightcurve[0]
    lightcurve_extended = np.concatenate((leading_edge_extension, lightcurve))

    new_lightcurve = np.convolve(sample, lightcurve_extended, 'valid')
    assert len(new_lightcurve) == len(lightcurve)

    # Now perform a left shift to make our convolution consistent with our start-of-exposure convention.
    # shift_count = len(sample) - 1
    shift_count = shift_needed
    lightcurve_shifted = np.roll(new_lightcurve, -shift_count)  # Do a 'left-roll'
    # Fix the points at the right edge that got overwritten by the 'roll'  Here we assume that the
    # lightcurve had already achieved 'steady state' at the end.
    for i in range(1, shift_count + 1):
        lightcurve_shifted[-i] = new_lightcurve[-1]

    return lightcurve_shifted


def sin_degrees(angle_degrees):
    # Convert angle from degrees to radians
    radians_per_degree = np.pi / 180
    angle_radians = angle_degrees * radians_per_degree
    return np.sin(angle_radians)


def get_star_chord_samples(
        star_diameter_mas: float,
        distance_to_asteroid_AU: float,
        shadow_speed: float,
        time_resolution: float,
        d_limb_angle_degrees: float = 90.0,
        r_limb_angle_degrees: float = 90.0,
        num_points_in_chord_base: int = 0
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Computes the array that can be convolved with the diffraction curve to incorporate star diameter effects.

    :param star_diameter_mas:        star disk diameter (units: milli-arcseconds)
    :param distance_to_asteroid_AU:  distance to asteroid (units: Astronomical Units)
    :param shadow_speed: asteroid speed (units: km/sec)
    :param time_resolution: resolution for chord slices (units: sec)
    :param d_limb_angle_degrees: limb angle at D (90 degrees = head-on)
    :param r_limb_angle_degrees: limb angle at R (90 degrees = head-on)
    :param num_points_in_chord_base: number of points in x axis of main plot
    :return: np.array of chord profiles, normalized so that sum(chords) = 1.0
    """
    star_diameter_radians = star_diameter_mas * 4.84814e-9
    distance_to_asteroid_km = distance_to_asteroid_AU * 149.6e6

    # print(star_diameter_radians, np.tan(star_diameter_radians), np.sin(star_diameter_radians))
    star_projection_km = 2 * np.tan(star_diameter_radians / 2) * distance_to_asteroid_km
    star_diameter_sec = star_projection_km / shadow_speed

    n_star_chords_in_radius = int(star_diameter_sec / 2 / time_resolution)
    print(f'time_resolution: {time_resolution:0.4f}  n_star_chords: {n_star_chords_in_radius}')
    n_r_limb_chords_in_radius = int(n_star_chords_in_radius / sin_degrees(r_limb_angle_degrees))
    n_d_limb_chords_in_radius = int(n_star_chords_in_radius / sin_degrees(d_limb_angle_degrees))

    print(f'n_d_limb_chords_in_radius: {n_d_limb_chords_in_radius}  '
          f'n_r_limb_chords_in_radius: {n_r_limb_chords_in_radius}')

    d_time_resolution = time_resolution * sin_degrees(d_limb_angle_degrees)
    r_time_resolution = time_resolution * sin_degrees(r_limb_angle_degrees)

    radius_sec = star_diameter_sec / 2
    r2 = radius_sec * radius_sec

    print(f'radius_sec**2: {r2:0.4f}')

    d_star_chords = np.zeros(num_points_in_chord_base)
    d_star_chords_standalone = np.zeros(n_d_limb_chords_in_radius * 2 + 1)
    normalizer = 0.0
    plot_margin = 20
    j = 0
    for i in range(-n_d_limb_chords_in_radius, n_d_limb_chords_in_radius + 1):
        chord = 2.0 * np.sqrt(r2 - (i * d_time_resolution) ** 2)
        if j == 0 or j == n_d_limb_chords_in_radius * 2:
            chord = 0
        d_star_chords[j + plot_margin] = chord
        d_star_chords_standalone[j] = chord
        j += 1
        normalizer += chord

    normed_d_chords = [c / normalizer for c in d_star_chords]
    normed_d_chords_standalone = [c / normalizer for c in d_star_chords_standalone]

    r_star_chords = np.zeros(num_points_in_chord_base)
    r_star_chords_standalone = np.zeros(n_r_limb_chords_in_radius * 2 + 1)

    normalizer = 0.0
    plot_offset = len(r_star_chords) - n_r_limb_chords_in_radius * 2 - plot_margin
    j = 0
    for i in range(-n_r_limb_chords_in_radius, n_r_limb_chords_in_radius + 1):
        chord = 2.0 * np.sqrt(r2 - (i * r_time_resolution) ** 2)
        if j == 0 or j == n_r_limb_chords_in_radius * 2:
            chord = 0.0
        r_star_chords[j + plot_offset] = chord
        r_star_chords_standalone[j] = chord
        j += 1
        normalizer += chord

    normed_r_chords = [c / normalizer for c in r_star_chords]
    normed_r_chords_standalone = [c / normalizer for c in r_star_chords_standalone]

    return np.array(normed_d_chords), np.array(normed_r_chords), \
        np.array(normed_d_chords_standalone), np.array(normed_r_chords_standalone)


def generate_underlying_lightcurve_plots(
        b_value=100.0,
        a_value=0.0,
        frame_time=None,
        ast_dist=None,
        shadow_speed=None,
        ast_diam=None,
        centerline_offset=None,
        star_diam=None,
        d_angle=None,
        r_angle=None,
        skip_central_flash=False,
        title_addon=''
):
    mid = (b_value + a_value) / 2

    if frame_time > 0.001:
        data_summary = f'\nframe time(sec): {frame_time:0.4f} '
    else:
        data_summary = '\n'
    if ast_dist is not None:
        data_summary += f'  asteroid distance(AU): {ast_dist:0.2f}'
    if shadow_speed is not None:
        data_summary += f'  shadow speed(km/sec): {shadow_speed:0.2f}'
    extra_title = title_addon

    ans = generate_transition_point_time_correction_look_up_tables(
        baseline_intensity=b_value,
        event_intensity=a_value,
        frame_time_sec=frame_time,
        asteroid_distance_AU=ast_dist,
        asteroid_diameter_km=ast_diam,
        shadow_speed_km_per_sec=shadow_speed,
        centerline_offset=centerline_offset,
        star_diameter_mas=star_diam,
        d_limb_angle_degrees=d_angle,
        r_limb_angle_degrees=r_angle,
        skip_central_flash=skip_central_flash
    )
    fig = plt.figure('Dplot', figsize=(10, 6), tight_layout=True)
    ax = fig.add_subplot(111)
    ax.set(xlabel='seconds', ylabel='Intensity')
    if star_diam is not None:
        star_comment = f'\nstar diam(mas): {star_diam:0.2f}  D limb angle: {d_angle:0.1f}' \
                       f'  R limb angle: {r_angle:0.1f}'
    else:
        star_comment = ''
    if ast_diam is not None and centerline_offset is not None:
        graze_comment = f'\nast diam(km): {ast_diam:0.2f} centerline offset(km): {centerline_offset:0.2f}'
    else:
        graze_comment = ''
    ax.set_title(extra_title + 'Lightcurve composition' + data_summary + star_comment + graze_comment)
    if ans['star_chords_d'] is not None:
        star_chords_d = ans['star_chords_d']
        rescaled_star_chords_d = star_chords_d * (b_value - a_value) / max(star_chords_d) / 2
        rescaled_star_chords_d += a_value
        n_star_chords = len(rescaled_star_chords_d)
        ax.plot(ans['time deltas'][:n_star_chords],
                rescaled_star_chords_d, label='star disk function', color='orange')

        star_chords_r = ans['star_chords_r']
        rescaled_star_chords_r = star_chords_r * (b_value - a_value) / max(star_chords_r) / 2
        rescaled_star_chords_r += a_value
        n_star_chords = len(rescaled_star_chords_r)
        ax.plot(ans['time deltas'][:n_star_chords],
                rescaled_star_chords_r, color='orange')

    right_edge_shadow = (ast_diam / 2) / shadow_speed
    ax.axvline(right_edge_shadow, linestyle='--', label='geometrical shadow', color='black')
    ax.axvline(-right_edge_shadow, linestyle='--', color='black')
    if frame_time > 0.001:
        offset = ans['time deltas'][-1] / 2
        ax.plot([offset, offset, offset + frame_time, offset + frame_time],
                [a_value, mid, mid, a_value], label='camera exposure function', color='red')
    # if ans['graze D'] is not None:
    #     ax.plot(ans['time deltas'], ans['graze D'], label='diffraction (grazed) D')
    # if ans['star D'] is None:
    #     ax.plot(ans['time deltas'], ans['raw D'], label='underlying lightcurve', color='green')
    # else:
    ax.plot(ans['time deltas'], ans['scaled underlying'], label='underlying lightcurve', color='green')

    if frame_time > 0.001:
        ax.plot(ans['time deltas'], ans['camera response'], label='camera response', color='red')

    plt.grid()
    ax.legend()
    main_fig = fig

    return main_fig, ans


def demo():
    main_figure = None

    tests_to_run = [3]
    print(f'=== tests to be run: {tests_to_run}')

    if 1 in tests_to_run:
        print(f'running test 1')
        lightcurve = np.concatenate((np.zeros(3), np.ones(2), np.zeros(2)))
        sample = np.ones(2)
        ans = lightcurve_convolve(sample=sample, lightcurve=lightcurve, shift_needed=len(sample) - 1)
        assert np.array_equal(ans, np.array([0, 0, 1, 2, 1, 0, 0]))

        fig = plt.figure(figsize=(10, 6))
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])

        ax.plot(ans, 'v', label='convolution')
        ax.plot(lightcurve, '^', label='lightcurve')
        ax.plot(sample, 'x', label='sampler')
        ax.legend()
        plt.show()

    if 2 in tests_to_run:
        print(f'running test 2')
        ans = generate_transition_point_time_correction_look_up_tables(
            baseline_intensity=10.0,
            event_intensity=-1.5, frame_time_sec=0.334
        )

        fig = plt.figure(figsize=(10, 6))
        ax = fig.add_subplot(111)
        ax.plot(ans['time deltas'], ans['D curve'], '.', label='D curve')
        ax.legend()

        fig = plt.figure(figsize=(10, 6))
        ax = fig.add_subplot(111)
        ax.plot(ans['time deltas'], ans['R curve'], '.', label='R curve')
        ax.legend()

    if 3 in tests_to_run:
        print(f'running test 3')
        b_value = 100.0
        a_value = 0.0
        skip_central_flash = True  # This speeds up computation when the central flash is not needed
        # frame_time = 0.001 # Will show underlying diffraction curve
        frame_time = 0.08
        # frame_time = 0.001
        star_diam = 14.2  # mas
        # star_diam = None
        d_angle = 75
        r_angle = 90
        # ast_dist = 2.5752  # Felicia
        ast_dist = 1.978
        ast_diam = 45
        centerline_offset = 0.0
        shadow_speed = 6.49

        decide_model_to_use(asteroid_diameter_km=ast_diam,
                            asteroid_distance_AU=ast_dist,
                            star_diameter_mas=star_diam,
                            shadow_speed=shadow_speed,
                            frame_time=frame_time)

        title_addon = '(Roma)  '

        main_figure, _ = generate_underlying_lightcurve_plots(
            b_value=100.0,
            a_value=0.0,
            frame_time=frame_time,
            ast_dist=ast_dist,
            ast_diam=ast_diam,
            centerline_offset=centerline_offset,
            shadow_speed=shadow_speed,
            star_diam=star_diam,
            d_angle=d_angle,
            r_angle=r_angle,
            title_addon=title_addon,
            skip_central_flash=skip_central_flash
        )

        _ = generate_transition_point_time_correction_look_up_tables(
            baseline_intensity=b_value,
            event_intensity=a_value,
            frame_time_sec=frame_time,
            asteroid_distance_AU=ast_dist,
            asteroid_diameter_km=ast_diam,
            centerline_offset=centerline_offset,
            shadow_speed_km_per_sec=shadow_speed,
            star_diameter_mas=star_diam,
            d_limb_angle_degrees=d_angle,
            r_limb_angle_degrees=r_angle,
            skip_central_flash=skip_central_flash
        )

    print(f'=== end tests')

    return main_figure


if __name__ == "__main__":
    # print(plt.get_backend())
    # plt.switch_backend('Qt5agg')
    # print(plt.get_backend())
    main_plot = demo()
    matplotlib.pyplot.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jun  3 09:37:59 2017

@author: Bob Anderson
"""

from numba import njit, float64, int64

from typing import List, Iterator, Union, Tuple
from scipy.signal import savgol_filter as savgol
from pyoteapp.solverUtils import model
from pyoteapp import autocorrtools
import numpy as np

# import pyximport
# pyximport.install()
# from pyoteapp.c_functions import find_Dedge_logl  # Finds D using a subframe model
# from pyoteapp.c_functions import wrap_logl


@njit(cache=True)
def loglikelihood(y: float64, m: float64, sigma: float64) -> float64:
    """ calculate ln(likelihood) given Gaussian statistics

    Args:
        y (double):     measured value
        m (double):     mean (expected model value)
        sigma (double): stdev of measurements

    Returns:
        natural logarithm of un-normalized probability based on Gaussian distribution
    """

    # -log(sqrt(2*pi)) = -0.9189385332046727

    return -0.9189385332046727 - np.log(sigma) - (y - m) * (y - m) / (2.0 * sigma * sigma)


# The following routine finds an D edge using a subframe model
@njit(float64(int64, float64[:], float64[:], float64[:], float64[:], float64, float64, float64, float64), cache=True)
def find_Dedge_logl_numba(n, y, mb, ma, mm, b, a, sigma_b, sigma_a):

    mcur = 0.0

    for i in range(n):
        mb[i] = loglikelihood(y[i], b, sigma_b)
        ma[i] = loglikelihood(y[i], a, sigma_a)
        if y[i] > b:
            mm[i] = mb[i]
        elif y[i] < a:
            mm[i] = ma[i]
        else:
            noise = sigma_b - ((b - y[i]) / (b - a)) * (sigma_b - sigma_a)
            mm[i] = loglikelihood(y[i], y[i], noise)
        mcur += ma[i]

    mcur = mcur - ma[0] + mm[0]
    edge_pos = 0

    mmax = mcur
    for i in range(n - 1):
        mcur = mcur - mm[i] + mb[i] - ma[i + 1] + mm[i + 1]
        if mcur > mmax:
            mmax = mcur
            edge_pos = i + 1

    if a <= y[edge_pos] <= b:
        adj = 1.0 - ((b - y[edge_pos]) / (b - a))
    elif y[edge_pos] > b:
        adj = 1.0
    else:  # y[edge_pos] < a
        adj = 0.0

    return float(edge_pos) + adj


def edgeDistributionGenerator(*, ntrials: int = 10000, numPts: int = None, D: int = None, acfcoeffs: List[float] = None,
                              B: float = None, A: float = None, sigmaB: float = None, sigmaA: float = None
                              ) -> Iterator[Union[float, List[float]]]:
    # my_noise_gen = None
    try:
        my_noise_gen = autocorrtools.CorrelatedNoiseGenerator(acfcoeffs)
    except np.linalg.LinAlgError:
        # New in version 2.0.7  Fall back to uncorrelated noise if Cholesky
        # fails.
        my_noise_gen = autocorrtools.CorrelatedNoiseGenerator([1.0])
        yield -1.0  # This is a flag value that is only returned when a LinAlgError exception occurs

    mb = np.ndarray(shape=(numPts,), dtype=np.double)  # 'model' B value
    ma = np.ndarray(shape=(numPts,), dtype=np.double)  # 'model' A value
    mm = np.ndarray(shape=(numPts,), dtype=np.double)  # 'model' intermediate value

    edgePos = np.zeros(shape=ntrials, dtype=np.float)

    m, sigma = model(B=B, A=A, D=D, R=-1, sigmaB=sigmaB, sigmaA=sigmaA, numPts=numPts)
    # m[:D] == B  m[D:] == A  sigma[:D] == sigmaB  sigma[:D] == sigmaA

    for i in range(ntrials):
        if i % 2000 == 0:
            yield i / ntrials
        noise = my_noise_gen.corr_normal(numPts)    
        y = m + noise * sigma
        
        centerValue = np.random.uniform(A, B)
        spanFrac = (B - centerValue) / (B - A)
        noise = sigmaB - spanFrac * (sigmaB - sigmaA)
        y[D] = np.random.normal(centerValue, noise)
        deltaPos = 1 - spanFrac
        edge_pos = find_Dedge_logl_numba(numPts, y, mb, ma, mm, B, A, sigmaB, sigmaA)

        edgePos[i] = edge_pos - deltaPos
        
    yield edgePos


def ciBars(*, dist: np.ndarray = None, ci: float = None, D: float = None) -> Tuple[float, ...]:
    assert(ci >= 0)

    # ecdfX means estimated cumulative distribution function
    # ecdfX = np.sort(dist)
    # ecdfY = np.arange(ecdfX.size) / ecdfX.size

    sortedDist = np.sort(dist)

    # Calculate all possible containment intervals that contain the fraction of solutions given by ci
    i0 = 0                          # left edge of containment interval
    i1 = int(ci * sortedDist.size)  # right edge of containment interval
    ciVec = np.ndarray(sortedDist.size-i1, dtype='float')
    while i1 < sortedDist.size:
        ciVec[i0] = sortedDist[i1] - sortedDist[i0]
        # Slide the containment interval one step to the right. By construction, this interval
        # still contains the same number of solutions
        i0 += 1
        i1 += 1

    # We choose the smallest containment interval for our error bars
    minCi = np.min(ciVec)
    minCiIndex = np.where(ciVec == minCi)[0][0]
    loBar = sortedDist[minCiIndex]
    midBar = D
    hiBar = loBar + minCi
    
    # indices = np.where(ecdfY >= 0.5 + ci / 2)[0]
    # if indices.size == 0:
    #     hiIndex = ecdfX.size
    # else:
    #     hiIndex = np.where(ecdfY >= 0.5 + ci / 2)[0][0]
    # hiBar = ecdfX[hiIndex]

    # indices = np.where(ecdfY >= 0.5 - ci / 2)[0]
    # if indices.size == 0:
    #     loIndex = 0
    # else:
    #     loIndex = np.where(ecdfY >= 0.5 - ci / 2)[0][0]
    # loBar = ecdfX[loIndex]

    # midBar = ecdfX[np.where(ecdfY >= 0.5)[0][0]]

    # varCi = np.var(ecdfX[loIndex:hiIndex+1])
    # meanCi = np.mean(ecdfX[loIndex:hiIndex+1])
    # print(f'ci: {ci:0.4f}  meanCi: {meanCi:0.4f}  varCi: {varCi:0.4f}  lo: {ecdfX[loIndex]:0.4f}  hi: {ecdfX[hiIndex]:0.4f}')
    # print(f'   sqrt(var)/ci: {np.sqrt(varCi)/ (hiBar - loBar):0.4f}')
    # totalVar = np.var(ecdfX)
    # print(f'   totalVar: {totalVar:0.4f}')
    
    return loBar, midBar, hiBar, loBar - midBar, hiBar - midBar


def sampledCdf(dist: np.ndarray, nsamples: int) -> Tuple[np.ndarray, np.ndarray]:
    sdist = np.sort(dist)
    
    # spt stands for 'sample point'
    spt = sdist[0]
    
    delta = (sdist[-1] - sdist[0]) / nsamples
    
    y = []
    x = []

    i = 0
    while spt <= sdist[-1]:
        x.append(spt)
        y.append(i/sdist.size)
        spt += delta
        while spt > sdist[i]:
            i += 1
            if i == sdist.size:
                break
           
    return np.array(x), np.array(y)


def smoothedSolutionDistribution(dist: np.ndarray, nsamples: int) -> Tuple[np.ndarray, np.ndarray]:
    sampledX, sampledY = sampledCdf(dist, nsamples)
    smoothedY = savgol(sampledY, 21, 3)
    
    return np.array(sampledX[:-1]), np.diff(smoothedY)


def createDurDistribution(dist: np.ndarray) -> np.ndarray:
    maxIndex = dist.size - 1
    
    def sample():
        return dist[np.random.randint(0, maxIndex)]
    
    durDist = np.ndarray(shape=(dist.size,))
    
    for i in range(dist.size):
        durDist[i] = sample() - sample()
        
    return durDist


def bob():
    # logl = wrap_logl(0.0, 1.0, 0.5)
    # print(f'logl: {logl}')
    logl = loglikelihood(0.0, 1.0, 0.5)
    print(f'logl: {logl}')

    n = 5
    y = np.array([5.1, 4.9, 0.9, 1.1, 1.0])
    mb = np.zeros_like(y)
    ma = np.zeros_like(y)
    mm = np.zeros_like(y)
    sigma_b = 1.0
    sigma_a = 0.5

    b = 5.0
    a = 1.0

    d_edge = find_Dedge_logl_numba(n, y, mb, ma, mm, b, a, sigma_b, sigma_a)
    print(f'd_edge (numba): {d_edge}')

    # d_edge = find_Dedge_logl(n, y, mb, ma, mm, b, a, sigma_b, sigma_a)
    # print(f'd_edge (c code): {d_edge}')


if __name__ == '__main__':
    bob()



import numpy as np


# Return the standard deviation of a single integration block
# that starts at 'index'.   'index' is assumed to point to a value in y[], so
# we also check for a 'block' that goes past the end of y[]
def std_of_block(index, blockSize, y):
    if index + blockSize > len(y):
        return False, 0
    else:
        return True, np.std(y[index:(index+blockSize)])


# Return the mean of the standard deviations of all integration blocks in
# y[], starting at 'index' (assumed in 0...blockSize-1)
def mean_stds(index, blockSize, y):
    ans = 0
    n = 0
    i = index
    while True:
        ok, std = std_of_block(i, blockSize, y)
        if not ok:
            break
        else:
            ans += std
            n += 1
        i = i + blockSize

    if n > 0:
        return ans / n
    else:
        return ans


# Returns a list of the block standard deviations for all possible
# 'index' (offset) values
#
# np.argmin(ans) will get the 'best' offset value while
# np.min(ans)    will get the corresponding 'best' standard deviation
def mean_std_versus_offset(blockSize, y):
    ans = []
    for i in range(0, blockSize):  # i is the offset of the first block
        ans.append(mean_stds(i, blockSize, y))
    return ans



r"""
This module provides several tools involving correlated noise.

There is a function for calculating auto-correlation coefficients
and a class for use in generating random noise with specified auto-correlation
characteristics.

Example:
    >>> noise_gen = CorrelatedNoiseGenerator(acfcoeffs=[1.0, 0.5, 0.3, 0.1])
    >>> test_noise = noise_gen.corr_normal(num_values=100000)
    >>> autocorr(test_noise, lastlag=6)
    [1.0,
     0.50094352543708198,
     0.30158412855919509,
     0.10104967953428118,
     0.002665198305403594,
     0.0024066568197502591,
     0.0012079072002819918]
"""

# coding: utf-8

from numpy import zeros, random, convolve
from scipy.stats.stats import pearsonr
from scipy.linalg import cholesky

__all__ = ['CorrelatedNoiseGenerator', 'autocorr']


def selfcorr(x, lag=1):
    """ computes the self-correlation of a vector for a specified lag.
        This is a support function for autocorr.
    
        Args:
            x (float array): array to be correlated
            lag (int): specifies the lag for the correlation calculation
        
        Returns:
            a single correlation coefficient at the specified lag.
    """

    n = len(x)

    if len(x) <= lag + 1:
        raise Exception("lag is too large for x array")

    y1 = x[lag:]
    y2 = x[:n - lag]
    corr = pearsonr(y1, y2)[0]
    return corr


def autocorr(x, lastlag=10):
    """ computes the self-correlation coefficients of x 
        from lag = 0 to lag = lastlag
    
        Args:
            x (float array): array to be correlated
            lastlag (int): the final lag value to be calculated
    """

    if lastlag + 1 >= len(x):
        raise Exception("laglast is too large")

    lags = range(0, lastlag + 1)
    corrs = [selfcorr(x, lag) for lag in lags]
    return corrs

# Our goal is to produce gaussian noise that has a specified auto-correlation response when given as an
# argument to ```autocorr()```. To do that, we build an auto-correlation matrix that has the desired
# auto-correlation coefficients in the off-diagonal terms.


def build_acf_matrix(acfcoeffs, size):
    """ builds a normalized auto-correlation matrix
        that has dimensions of size*size.
        
        Args:
            acfcoeffs (float array): the desired auto-correlation coefficients
            size (int): the size of the ultimate noise array
    """

    ans = zeros((size, size))
    k = len(acfcoeffs)
    m = min(k, size)  # Just in case k > size

    for i in range(m):
        for j in range(size - i):
            ans[j, j + i] = acfcoeffs[i]
            ans[j + i, j] = acfcoeffs[i]

    return ans


class CorrelatedNoiseGenerator:
    """ This correlated noise generator class takes advantage of the fact that the entries in the cholesky
        decomposition of the acf matrix rapidly approach limiting values.  This means that we do not need to build
        a large acf matrix: instead we make it large enough that the entries in the cholesky matrix have reached
        limiting values. We extract those limiting values along with a matching set of initial random numbers.
    """

    def __init__(self, acfcoeffs):
        """ Uses acfcoeffs and cholesky decomposition to calculate final_coeffs, the
            limiting values in the cholesky matrix columns. There will be as many such
            values as there are entries in acfcoeffs.

            An initial set of N[0,1] random numbers are generated and stored in r[].
            With this setup completed, correlated random number can be generated by:

                r[] convolve final_coeffs[]
        """

        if len(acfcoeffs) < 1:
            raise Exception("Invalid acfcoeffs array.")
        if acfcoeffs[0] != 1.0:
            raise Exception("Invalid acfcoeffs list; first entry not equal to 1.0")

        self.acf_length = len(acfcoeffs)

        # We only make the acf_matrix big enough to allow the subsequent cholesky
        # decomposition to reach limiting values. A size of 10 times the number of
        # acfcoeffs is more than adequate.
        self.acfmatrix = build_acf_matrix(acfcoeffs, size=self.acf_length * 10)

        self.noise_matrix = cholesky(self.acfmatrix, lower=False)
        # The noise_matrix is composed mostly of zeroes. For each column, there
        # will be as many non-zero entries as there are in acfcoeffs. The limiting
        # values are in the bottom of the last column.
        self.final_coeffs = self.noise_matrix[-self.acf_length:, -1]

    def corr_normal(self, num_values):
        """ Calculates and returns a NumPy ndarray of length
            num_values. The values are normally distributed with a mean of
            zero, a standard deviation of 1, and an auto-correlation that
            closely approximates the values given by acfcoeffs when this
            object was instantiated.

            Args:
                num_values (int): the desired length of the output array
        """

        return convolve(self.final_coeffs, random.normal(size=num_values), 'same')



import numpy as np
import warnings

# suppress warnings
warnings.filterwarnings('ignore')


def expRedge(B, A, n, offset, N):
    vstart = A
    vend = B
    n -= offset
    vout = vstart - (vstart - vend) * (1.0 - np.exp(-n / N))
    if vend < vstart < vout:
        vout = vstart
    if vend > vstart > vout:
        vout = vstart
    return vout


def expDedge(B, A, n, offset, N):
    vstart = B
    vend = A
    n -= offset
    vout = vstart - (vstart - vend) * (1.0 - np.exp(-n / N))
    if vend < vstart < vout:
        vout = vstart
    if vend > vstart > vout:
        vout = vstart
    return vout


def getRedgePoints(numPoints, B, A, offset, N):
    # Returns 2 * numPoints + 1 values

    # Generate points to the left of the 'edge'
    edgePoints = [A for _ in range(numPoints)]

    # Now generate numPoints + 1 to the right of the 'edge'
    n = 0
    for i in range(numPoints + 1):
        vn = expRedge(B, A, n, offset, N)
        edgePoints.append(vn)
        n += 1
    return edgePoints


def getDedgePoints(numPoints, B, A, offset, N):
    # Returns 2 * numPoints + 1 values

    # Generate points to the left of the 'edge'
    edgePoints = [B for _ in range(numPoints)]

    # Now generate numPoints + 1 to the right of the 'edge'
    n = 0
    for i in range(numPoints + 1):
        vn = expDedge(B, A, n, offset, N)
        edgePoints.append(vn)
        n += 1
    return edgePoints


def scoreDedge(numTheoryPts, B, A, offset, N, actual, matchPoint):
    theory = getDedgePoints(numTheoryPts, B, A, offset, N)
    nTheory = len(theory)
    assert len(actual) >= nTheory
    assert matchPoint + numTheoryPts <= len(actual)
    metric = 0.0
    for i in range(nTheory):
        metric += (theory[i] - actual[i + matchPoint]) ** 2
    return metric


def scoreRedge(numTheoryPts, B, A, offset, N, actual, matchPoint):
    theory = getRedgePoints(numTheoryPts, B, A, offset, N)
    nTheory = len(theory)
    assert len(actual) >= nTheory
    assert matchPoint + numTheoryPts <= len(actual)
    metric = 0.0
    for i in range(nTheory):
        metric += (theory[i] - actual[i + matchPoint]) ** 2
    return metric


def locateIndexOfBestMatchPoint(numTheoryPts, B, A, offset, N, actual, edge='D'):
    metricVec = []
    for k in range(len(actual) - numTheoryPts * 2):
        if edge == 'D':
            metricVec.append(scoreDedge(numTheoryPts, B, A, offset, N, actual, k))
        else:
            metricVec.append(scoreRedge(numTheoryPts, B, A, offset, N, actual, k))

    return np.where(metricVec == np.min(metricVec))[0][0]


def locateBestOffset(numTheoryPts, B, A, N0, actual, matchPoint, edge='D'):
    metricVec = []
    offsetVec = []
    for offset in np.linspace(-2, 2, 81):
        if edge == 'D':
            metricVec.append(scoreDedge(numTheoryPts, B, A, offset, N0, actual, matchPoint))
        else:
            metricVec.append(scoreRedge(numTheoryPts, B, A, offset, N0, actual, matchPoint))
        offsetVec.append(offset)

    indexToBestOffset = np.where(metricVec == np.min(metricVec))[0][0]
    return offsetVec[indexToBestOffset]


def locateBestN(numTheoryPts, B, A, offset, N0, actual, matchPoint, edge='D'):
    metricVec = []
    NVec = []
    if N0 - 1 <= 0:
        lowN = 0.1
        hiN = lowN + 2.0
    else:
        lowN = N0 - 1.0
        hiN = lowN + 1.0
    for N in np.linspace(lowN, hiN, 41):
        if edge == 'D':
            metricVec.append(scoreDedge(numTheoryPts, B, A, offset, N, actual, matchPoint))
        else:
            metricVec.append(scoreRedge(numTheoryPts, B, A, offset, N, actual, matchPoint))
        NVec.append(N)

    indexToBestN = np.where(metricVec == np.min(metricVec))[0][0]
    return NVec[indexToBestN]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 29 07:00:33 2017

@author: Bob Anderson
"""
MIN_SIGMA = 0.1

from math import exp
import numpy as np
from pyoteapp.likelihood_calculations import cum_loglikelihood, aicc, logLikelihoodLine
from pyoteapp.likelihood_calculations import loglikelihood
from numba import njit


def aicModelValue(*, obsValue=None, B=None, A=None, sigmaB=None, sigmaA=None):
    assert(B >= A)
    assert(sigmaA > 0.0)
    assert(sigmaB > 0.0)
    # This function determines if an observation point should categorized as a baseline (B)
    # point, an event (A) point, or a valid intermediate point using the Akaike Information Criterion
    # An intermediate point reflects a more complex model (higher dimension model)
    if obsValue >= B:
        return B  # Categorize as baseline point
    if obsValue <= A:
        return A  # Categorize as event point
    if B == A:
        return B

    sigmaM = sigmaA + (sigmaB - sigmaA) * ((obsValue - A) / (B - A))
    loglB = loglikelihood(obsValue, B, sigmaB)
    loglM = loglikelihood(B, B, sigmaM) - 1.0  # The -1 is the aic model complexity 'penalty'
    loglA = loglikelihood(obsValue, A, sigmaA)

    if loglM > loglA and loglM > loglB:
        return obsValue  # Categorize as valid intermediate value
    elif loglB > loglA:
        return B  # Categorize as baseline point
    else:
        return A  # Categorize as event point


@njit(cache=True)
def model(B=0.0, A=0.0,
          D=-1, R=-1,
          sigmaB=0.0, sigmaA=0.0, numPts=0):
    
    # D, R = edgeTuple
    assert(numPts > 0 and sigmaA >= 0 and sigmaB >= 0)
    # TODOnot Remove this experimental code and uncomment the line above
    # if sigmaA < 0:
    #     sigmaA = 0.0
    # if sigmaB < 0:
    #     sigmaB = 0.0
    # assert(numPts > 0)
    # # assert(sigmaA >= 0.0)
    # # assert(sigmaB >= 0.0)
    m = np.zeros(numPts)
    sigma = np.zeros(numPts)

    if D == -1 and R == -1:  # -1 means None
        m[:] = B
        sigma[:] = sigmaB
        
        return m, sigma
    
    if not D == -1 and not R == -1:  # D and R are not None
        assert((D >= 0) and (R > D))
        assert((D < numPts) and (R < numPts))
        assert(R > D)
        
        m[:D] = B
        m[D:R] = A
        m[R:] = B
        
        sigma[:D] = sigmaB
        sigma[D:R] = sigmaA
        sigma[R:] = sigmaB
        
        return m, sigma
    
    if not D == -1 and R == -1:  # R is None
        assert((D >= 0) and (D < numPts))
        
        m[:D] = B
        m[D:] = A
        
        sigma[:D] = sigmaB
        sigma[D:] = sigmaA
        
        return m, sigma
    
    if not R == -1 and D == -1:  # D is None
        assert((R >= 1) and (R < numPts))
        
        m[:R] = A
        m[R:] = B
        
        sigma[:R] = sigmaA
        sigma[R:] = sigmaB
        
        return m, sigma
            
    raise Exception("Unexpected condition.")
    

def candidatesFromEventSize(*, eventType='DandR',
                            left=None, right=None, 
                            minSize=None, maxSize=None):
    """
    This is implemented as a generator because if numPts is large, it is
    easy to get a candidate list in the many millions, so it's best to
    create the candidates one at a time.
    """

    assert(right > left)
    assert(minSize >= 1)
    assert(maxSize >= minSize)
    assert(maxSize < right - left)
    
    if eventType == 'Donly':
        for D in range(right-maxSize+1, right-minSize+2):
            yield D, None
    
    elif eventType == 'Ronly':
        for R in range(left+minSize, left+maxSize+1):
            yield None, R
    
    elif eventType == 'DandR':
        for size in range(minSize, maxSize+1):
            for pos in range(left+1, right+1-size):
                yield pos, pos+size
        
    else:
        raise Exception("Unrecognized event type")
        

def candidatesFromDandRlimits(*, eventType='DandR',
                              dLimits=None, rLimits=None):
    """
    This is implemented as a generator because if large limits are given, it is
    easy to get a very large candidate list, so it's best to
    create the candidates one at a time.
    """
    # The D and R limits are assumed valid as input and are non-overlapping
    
    if eventType == 'Donly':
        for D in range(dLimits[0], dLimits[1] + 1):
            yield D, None
    
    elif eventType == 'Ronly':
        for R in range(rLimits[0], rLimits[1] + 1):
            yield None, R
    
    elif eventType == 'DandR':
        for D in range(dLimits[0], dLimits[1] + 1):
            for R in range(rLimits[0], rLimits[1] + 1):
                yield D, R
        
    else:
        raise Exception("Unrecognized edvent type")
        

@njit(cache=True)
def calcNumCandidatesFromDandRlimits(eventType='DandR', d_start=-1, d_end=-1,
                                     r_start=-1, r_end=-1):
    
    # The D and R limits are assumed valid as input and are non-overlapping
    
    if eventType == 'Donly':
        return d_end - d_start + 1
    elif eventType == 'Ronly':
        return r_end - r_start + 1
    elif eventType == 'DandR':
        return (d_end - d_start + 1) * (r_end - r_start + 1)
    else:
        raise Exception("Unrecognized event type")


@njit(cache=True)
def calcNumCandidatesFromEventSize(eventType='DandR',
                                   left=None, right=None,
                                   minSize=None, maxSize=None):
    
    numPts = right - left + 1
    assert(numPts >= 0)
    assert(minSize >= 1)
    assert(maxSize >= minSize)
    assert(maxSize < numPts - 1)

    if eventType == 'DandR':
        c1 = maxSize - minSize + 1
        c2 = 2 * numPts - 2 - minSize - maxSize
        return int(c1 * c2 / 2)
    elif eventType == 'Donly':
        return maxSize - minSize + 1
    elif eventType == 'Ronly':
        return maxSize - minSize + 1
    else:
        raise Exception("Unrecognized edge specifier")


def newCalcBandA(*, yValues=None, tpList=None, left=None, right=None, cand=None):
    assert (right > left)

    sigmaA = 0.0

    def valuesToBeUsed(lowRange, highRange):
        valList = []
        for i in range(lowRange, highRange):
            if i not in tpList:
                valList.append(yValues[i])
        return valList

    D, R = cand  # Extract D and R from the tuple

    if R is None:
        assert (D >= left)
        # This is a 'Donly' candidate
        # Note that the yValue at D is not included in the B calculation
        # because that point is in the event bottom.
        valuesToUse = valuesToBeUsed(left, D)
        B = np.mean(valuesToUse)
        sigmaB = np.std(valuesToUse)
        # We have to deal with a D at the right edge.  There is no value to
        # the right to use to calculate A so we simply return the value at D
        # as the best estimate of A
        if D == right:
            A = yValues[D]
            sigmaA = 0.0
        else:
            # changed in 4.4.7
            # A = np.mean(yValues[D+1:right+1])
            valuesToUse = valuesToBeUsed(D, right + 1)
            A = np.mean(valuesToUse)
            sigmaA = np.std(valuesToUse)
        if A >= B:
            A = B * 0.999
        return B, A, sigmaB, sigmaA

    elif D is None:
        assert (R <= right)
        # This is an 'Ronly' candidate
        # We have to deal with a R at the right edge.  There is no value to
        # the right to use to calculate B so we simply return the value at R
        # as the best estimate of B
        if R == right:
            B = yValues[R]
            sigmaB = 0.0
        else:
            # changed in 4.4.7
            # B = np.mean(yValues[R+1:right+1])
            valuesToUse = [val for i, val in enumerate(yValues[R:right + 1]) if i not in tpList]
            B = np.mean(valuesToUse)
            sigmaB = np.std(valuesToUse)
        valuesToUse = valuesToBeUsed(left, R)
        A = np.mean(valuesToUse)
        sigmaA = np.std(valuesToUse)
        if A >= B:
            A = B * 0.999
        return B, A, sigmaB, sigmaA

    else:
        assert ((D >= left) and (R <= right) and (R > D))
        # We have a 'DandR' candidate
        leftBvals = valuesToBeUsed(left, D)

        if R == right:
            rightBvals = yValues[right]
        else:
            # changed in 4.4.8
            rightBvals = valuesToBeUsed(R, right + 1)
        B = (np.sum(leftBvals) + np.sum(rightBvals)) / (len(leftBvals) + len(rightBvals))
        sigmaB = np.std(leftBvals + rightBvals)

        if R - D == 1:  # Event size of 1 has no valid A --- we choose the value at D
            A = yValues[D]
        else:
            valuesToUse = valuesToBeUsed(D, R)
            A = np.mean(valuesToUse)
            sigmaA = np.std(valuesToUse)
        if A >= B:
            A = B * 0.999
        return B, A, sigmaB, sigmaA


def calcBandA(*, yValues=None, left=None, right=None, cand=None):
    
    assert(right > left)
    
    D, R = cand  # Extract D and R from the tuple
        
    if R is None:
        assert(D >= left)
        # This is a 'Donly' candidate
        # Note that the yValue at D is not included in the B calculation
        # because that point is in the event bottom.
        B = np.mean(yValues[left:D])
        # We have to deal with a D at the right edge.  There is no value to
        # the right to use to calculate A so we simply return the value at D
        # as the best estimate of A
        if D == right:
            A = yValues[D]
        else:
            # changed in 4.4.7
            # A = np.mean(yValues[D+1:right+1])
            A = np.mean(yValues[D:right+1])
        if A >= B:
            A = B * 0.999
        return B, A
    
    elif D is None:
        assert(R <= right)
        # This is an 'Ronly' candidate
        # We have to deal with a R at the right edge.  There is no value to
        # the right to use to calculate B so we simply return the value at R
        # as the best estimate of B
        if R == right:
            B = yValues[R]
        else:
            # changed in 4.4.7
            # B = np.mean(yValues[R+1:right+1])
            B = np.mean(yValues[R:right+1])
        A = np.mean(yValues[left:R])  # smallest R is left + 1
        if A >= B:
            A = B * 0.999
        return B, A
    
    else:
        assert((D >= left) and (R <= right) and (R > D))
        # We have a 'DandR' candidate
        leftBvals = yValues[left:D]  # Smallest D will = left
        if R == right:
            rightBvals = yValues[right]
        else:
            # changed in 4.4.8
            rightBvals = yValues[R:right+1]
        B = (np.sum(leftBvals) + np.sum(rightBvals)) / (leftBvals.size + rightBvals.size)

        if R - D == 1:  # Event size of 1 has no valid A --- we choose the value at D
            A = yValues[D]
        else:
            A = np.mean(yValues[D:R])
        if A >= B:
            A = B * 0.999
        return B, A


# def scoreCandidate(yValues, left, right, cand, sigmaB, sigmaA):
#     B, A = calcBandA(yValues=yValues, left=left, right=right, cand=cand)
#     m, sigma = model(B=B, A=A, D=cand[0], R=cand[1],
#                      sigmaB=sigmaB, sigmaA=sigmaA, numPts=yValues.size)
#     return cum_loglikelihood(yValues, m, sigma, left, right), B, A


def scoreSubFrame(yValues, left, right, cand, sigmaB, sigmaA):
    B, A = calcBandA(yValues=yValues, left=left, right=right, cand=cand)
    m, sigma = model(B=B, A=A, D=cand[0], R=cand[1],
                     sigmaB=sigmaB, sigmaA=sigmaA, numPts=yValues.size)
    D, R = cand
    if D is not None:
        if (yValues[D] < B) and (yValues[D] > A):
            m[D] = yValues[D]
    if R is not None:
        if (yValues[R] < B) and (yValues[R] > A):
            m[R] = yValues[R]
    return cum_loglikelihood(yValues, m, sigma, left, right), B, A
    

def candidateCounter(*, eventType='DandR',
                     dLimits=None, rLimits=None, 
                     left=None, right=None,
                     numPts=None, minSize=None, maxSize=None):

    def minMaxOk():
        if numPts is None:
            return False
        if minSize is None:
            return False
        if maxSize is None:
            return False
        if minSize < 1:
            return False
        if maxSize > (numPts - 2):
            return False
        if maxSize < minSize:
            return False
        return True
    
    # D and R limits trumps event size as candidate generator/counter
    if eventType == 'Donly':
        if dLimits:
            return ('usedLimits', calcNumCandidatesFromDandRlimits(eventType=eventType,
                    d_start=dLimits[0], d_end=dLimits[1], r_start=-1, r_end=-1))
        else:
            if minMaxOk():
                return ('usedSize', calcNumCandidatesFromEventSize(eventType=eventType, 
                        left=left, right=right, minSize=minSize, maxSize=maxSize))
            else:
                return 'error', -1
            
    elif eventType == 'Ronly':
        if rLimits:
            return ('usedLimits', calcNumCandidatesFromDandRlimits(eventType=eventType,
                    d_start=-1, d_end=-1,  r_start=rLimits[0], r_end=rLimits[1]))
        else:
            if minMaxOk():
                return ('usedSize', calcNumCandidatesFromEventSize(eventType=eventType, 
                        left=left, right=right, minSize=minSize, maxSize=maxSize))
            else:
                return 'error', -1
        
    elif eventType == 'DandR':
        if rLimits and dLimits:
            return ('usedLimits', calcNumCandidatesFromDandRlimits(eventType=eventType,
                    d_start=dLimits[0], d_end=dLimits[1], r_start=rLimits[0], r_end=rLimits[1]))
        else:
            if minMaxOk():
                return ('usedSize', calcNumCandidatesFromEventSize(eventType=eventType, 
                        left=left, right=right, minSize=minSize, maxSize=maxSize))
            else:
                return 'error', -1
            
    else:
        raise Exception("Unrecognized event type")


def subFrameAdjusted(*, eventType=None, cand=None, B=None, A=None,
                     sigmaA=None, sigmaB=None,
                     yValues=None, left=None, right=None):

    def adjustR():
        value = yValues[R]
        adj = (B - value) / (B - A)
        return R + adj

    def adjustD():
        value = yValues[D]
        adj = (value - A) / (B - A)
        return D + adj

    transitionPoints = []

    D, R = cand
    adjD = D
    adjR = R

    # Here we add code so we can analyze light curves that may have sigmaB or
    #  sigmaA values of zero.  This happens when testing with artificial data
    #  but can also result from real light curves that may be clipped so that
    #  all B pixels have a constant value.  Limovie can produce a sigmaA=0
    # when a rectangular aperture is in use as well

    if sigmaA == 0.0:
        sigmaA = MIN_SIGMA
    if sigmaB == 0.0:
        sigmaB = MIN_SIGMA

    if eventType == 'Donly':
        if aicModelValue(obsValue=yValues[D], B=B, A=A, sigmaB=sigmaB, sigmaA=sigmaA) == yValues[D]:
            # If point at D categorizes as M (valid mid-point), do sub-frame
            # adjustment and exit
            adjD = adjustD()
            transitionPoints.append(D)
            B, A, sigmaB, sigmaA = newCalcBandA(yValues=yValues, tpList=transitionPoints,
                                                left=left, right=right, cand=(D, R))
        elif aicModelValue(obsValue=yValues[D], B=B, A=A, sigmaB=sigmaB, sigmaA=sigmaA) == B:
            # else if point at D categorizes as B, set D to D+1 and recalculate B and A
            D = D + 1
            adjD = D
            B, A, sigmaB, sigmaA = newCalcBandA(yValues=yValues, tpList=transitionPoints,
                                                left=left, right=right, cand=(D, R))
            # It's possible that this new point qualifies as M --- so we check:
            if aicModelValue(
                    obsValue=yValues[D], B=B, A=A, sigmaB=sigmaB,
                    sigmaA=sigmaA) == yValues[D]:
                adjD = adjustD()
                transitionPoints.append(D)
                B, A, sigmaB, sigmaA = newCalcBandA(yValues=yValues, tpList=transitionPoints,
                                                    left=left, right=right, cand=(D, R))
        # else (point at D categorizes as A) --- nothing to do
        return [adjD, adjR], B, A, sigmaB, sigmaA

    elif eventType == 'Ronly':
        if aicModelValue(obsValue=yValues[R], B=B, A=A, sigmaB=sigmaB, sigmaA=sigmaA) == yValues[R]:
            # If point at R categorizes as M, do sub-frame adjustment
            adjR = adjustR()
            transitionPoints.append(R)
            B, A, sigmaB, sigmaA = newCalcBandA(yValues=yValues, tpList=transitionPoints,
                                                left=left, right=right, cand=(D, R))
        elif aicModelValue(obsValue=yValues[R], B=B, A=A, sigmaB=sigmaB, sigmaA=sigmaA) == A:
            # else if point at R categorizes as A, set R to R + 1 and recalculate B and A
            R = R + 1
            adjR = R
            B, A, sigmaB, sigmaA = newCalcBandA(yValues=yValues, tpList=transitionPoints,
                                                left=left, right=right, cand=(D, R))
            # It's possible that this new point qualifies as M --- so we check
            if aicModelValue(
                    obsValue=yValues[R], B=B, A=A, sigmaB=sigmaB,
                    sigmaA=sigmaA) == yValues[R]:
                adjR = adjustR()
                transitionPoints.append(R)
                B, A, sigmaB, sigmaA = newCalcBandA(yValues=yValues, tpList=transitionPoints,
                                                    left=left, right=right, cand=(D, R))
        elif aicModelValue(obsValue=yValues[R - 1], B=B, A=A, sigmaB=sigmaB,
                           sigmaA=sigmaA) == yValues[R - 1]:
            # The point at R categorizes as B, and we have found
            # that the point at R-1 categorizes as M, so set R to R-1 and
            # recalculate B and A
            R = R - 1
            adjR = adjustR()
            transitionPoints.append(R)
            B, A, sigmaB, sigmaA = newCalcBandA(yValues=yValues, tpList=transitionPoints,
                                                left=left, right=right, cand=(D, R))
        return [adjD, adjR], B, A, sigmaB, sigmaA

    elif eventType == 'DandR':
        if not R - D > 2:
            return [D, R], B, A, sigmaB, sigmaA
        if aicModelValue(
                obsValue=yValues[D], B=B, A=A, sigmaB=sigmaB, sigmaA=sigmaA) == yValues[D]:
            # The point at D categorizes as M, do sub-frame adjustment; this
            # (finishes D)
            adjD = adjustD()
            transitionPoints.append(D)
            B, A, sigmaB, sigmaA = newCalcBandA(yValues=yValues, tpList=transitionPoints,
                                                left=left, right=right, cand=(D, R))
        elif aicModelValue(obsValue=yValues[D], B=B, A=A, sigmaB=sigmaB, sigmaA=sigmaA) == B:
            # The point at D categorizes as B, set D to D+1 and recalculate B and A
            D = D + 1
            adjD = D
            # It's possible that this new point qualifies as M --- so we check
            if aicModelValue(
                    obsValue=yValues[D], B=B, A=A, sigmaB=sigmaB,
                    sigmaA=sigmaA) == yValues[D]:
                adjD = adjustD()
                transitionPoints.append(D)
                B, A, sigmaB, sigmaA = newCalcBandA(yValues=yValues, tpList=transitionPoints,
                                                    left=left, right=right, cand=(D, R))
        elif aicModelValue(obsValue=yValues[D-1], B=B, A=A, sigmaB=sigmaB,
                           sigmaA=sigmaA) == yValues[D-1]:
            # The point at D categorizes as A, and we have found
            # that the point at D-1 categorizes as M, so set D to D-1 and
            # recalculate B and A
            D = D - 1
            adjD = adjustD()
            transitionPoints.append(D)
            B, A, sigmaB, sigmaA = newCalcBandA(yValues=yValues, tpList=transitionPoints,
                                                left=left, right=right, cand=(D, R))

        if aicModelValue(
                obsValue=yValues[R], B=B, A=A, sigmaB=sigmaB, sigmaA=sigmaA) == yValues[R]:
            # The point at R categorizes as M, do sub-frame adjustment
            adjR = adjustR()
            transitionPoints.append(R)
            B, A, sigmaB, sigmaA = newCalcBandA(yValues=yValues, tpList=transitionPoints,
                                                left=left, right=right, cand=(D, R))
        elif aicModelValue(obsValue=yValues[R], B=B, A=A, sigmaB=sigmaB, sigmaA=sigmaA) == A:
            # The point at R categorizes as A, set R to R + 1 and recalculate B and A
            R = R + 1
            adjR = R
            B, A, sigmaB, sigmaA = newCalcBandA(yValues=yValues, tpList=transitionPoints,
                                                left=left, right=right, cand=(D, R))
            # It's possible that this new point qualifies as M --- so we check
            if aicModelValue(
                    obsValue=yValues[R], B=B, A=A, sigmaB=sigmaB,
                    sigmaA=sigmaA) == yValues[R]:
                adjR = adjustR()
                transitionPoints.append(R)
                B, A, sigmaB, sigmaA = newCalcBandA(yValues=yValues, tpList=transitionPoints,
                                                    left=left, right=right, cand=(D, R))
        elif aicModelValue(obsValue=yValues[R - 1], B=B, A=A, sigmaB=sigmaB,
                           sigmaA=sigmaA) == yValues[R - 1]:
            # The point at R categorizes as B, and we have found
            # that the point at R-1 categorizes as M, so set R to R-1 and
            # recalculate B and A
            R = R - 1
            adjR = adjustR()
            transitionPoints.append(R)
            B, A, sigmaB, sigmaA = newCalcBandA(yValues=yValues, tpList=transitionPoints,
                                                left=left, right=right, cand=(D, R))
        return [adjD, adjR], B, A, sigmaB, sigmaA

    else:
        raise Exception('Unrecognized event type')


def solver(*, eventType=None, yValues=None,
           left=None, right=None,
           sigmaB=None, sigmaA=None, 
           dLimits=None, rLimits=None,
           minSize=None, maxSize=None):
    
    bestCand = None
    bestScore = float('-inf')
    bestB = None
    bestA = None
    
    mode, numCandidates = candidateCounter(eventType=eventType, dLimits=dLimits, rLimits=rLimits,
                                           left=left, right=right, numPts=yValues.size,
                                           minSize=minSize, maxSize=maxSize)
    
    if mode == 'error':
        return bestCand, bestB, bestA
    
    if mode == 'usedLimits':
        candGen = candidatesFromDandRlimits(eventType=eventType, dLimits=dLimits,
                                            rLimits=rLimits)
    elif mode == 'usedSize':
        candGen = candidatesFromEventSize(eventType=eventType, left=left, right=right, 
                                          minSize=minSize, maxSize=maxSize)
    else:
        raise Exception('candidateCounter() returned unexpected "mode" ')
     
    counter = 0
    for cand in candGen:
        score, B, A = scoreSubFrame(yValues, left, right, cand, sigmaB, sigmaA)
        if score > bestScore:
            bestScore = score
            bestB = B
            bestA = A
            bestCand = cand
        counter += 1
        if counter % 1000 == 0:
            yield 'fractionDone', counter/numCandidates
           
    if eventType == 'DandR':
        k = 4
    else:
        k = 3
        
    # lineScore = logLikelihoodLine(yValues, sigmaB=sigmaB, left=left, right=right)
    lineScore = logLikelihoodLine(yValues, sigmaB=np.sqrt(np.var(yValues)), left=left, right=right)
    aiccSol = aicc(bestScore, right-left+1, k)
    aiccLine = aicc(lineScore, right-left+1, 1)
    if aiccSol < aiccLine:
        pLine = exp(-(aiccLine - aiccSol)/2)
    else:
        pLine = 1.00
    if pLine > 0.001:
        yield 'no event present', counter/numCandidates

    yield subFrameAdjusted(eventType=eventType, cand=bestCand, 
                           B=bestB, A=bestA, sigmaB=sigmaB, sigmaA=sigmaA,
                           yValues=yValues, left=left, right=right)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 23 07:24:21 2017

@author: Bob Anderson
"""

import os

tangraNeedsBackgroundSubtraction = True

pymovieSignalColumnCount = 0

pymovieDataColumns = []

pymovieColumnNamePrefix = 'signal'


def readLightCurve(filepath, pymovieColumnType='signal'):
    """
    Reads the intensities and timestamps from Limovie,
    Tangra, PYOTE, or R-OTE csv files.  (PYOTE and R-OTE file formats are equal)
    """
    global pymovieColumnNamePrefix

    pymovieColumnNamePrefix = pymovieColumnType
    if fileCanBeOpened(filepath):
        
        readOk, errMsg, frame, time, value, ref1, ref2, ref3, extra, aperture_names, headers = readAs(filepath)
        if readOk:
            return frame, time, value, ref1, ref2, ref3, extra, aperture_names, headers
        else:
            raise Exception(errMsg)
            
    else:
        raise Exception('File could not be opened')


def fileCanBeOpened(file):
    return os.path.exists(file) and os.path.isfile(file)
                   

def getFileKind(file):
    fileobject = open(file)
    with fileobject:
        line = fileobject.readline()
        if 'Tangra' in line:
            return 'Tangra'
        elif 'Limovie' in line:
            return 'Limovie'
        elif 'PyMovie' in line:
            return'PyMovie'
        elif 'R-OTE' in line or line[0] == '#':  # Matches PyOTE and PyMovie files too!
            return 'R-OTE'
        elif 'RAW' in line:
            return 'raw'
        else:
            return '???'


# noinspection PyUnusedLocal
def tangraParser(line, frame, time, value, ref1, ref2, ref3, extra):
    """
    We only accept Tangra files that have been formatted
    according to the AOTA default which is ---
       UT time field formatted as  [hh:mm:ss.sss]
       We detect the state of background subtraction (either done or needed)
       An example data line:  11,[16:00:14.183],2837.8,100.0,4097.32,200.0
    """
    part = line.split(',')
    if len(part) < 2:
        raise Exception(line + " :is an invalid Tangra file entry.")
    else:
        frame.append(part[0])
        time.append(part[1])

    try:
        for item in part:
            if item == '':
                raise Exception(line + " :cannot be parsed.  Are there empty fields in data lines? Fix them all!")

        if tangraNeedsBackgroundSubtraction:
            value.append(str(float(part[2]) - float(part[3])))
            if len(part) >= 6:
                if part[4]:
                    ref1.append(str(float(part[4]) - float(part[5])))
            if len(part) >= 8:
                if part[6]:
                    ref2.append(str(float(part[6]) - float(part[7])))
            if len(part) >= 10:
                if part[8]:
                    ref3.append(str(float(part[8]) - float(part[9])))
        else:
            value.append(part[2])
            if len(part) >= 4:
                if part[3]:
                    ref1.append(part[3])
            if len(part) >= 5:
                if part[4]:
                    ref2.append(part[4])
            if len(part) >= 6:
                if part[5]:
                    ref3.append(part[5])
    except ValueError:
        raise Exception(line + " :cannot be parsed.  Are there empty fields?")


# noinspection PyUnusedLocal
def limovieParser(line, frame, time, value, ref1, ref2, ref3, extra):
    """
    Limovie sample line ---
        3.5,21381195,21381200,22,27,43.0000,,,,,2737.8,3897.32 ...
    """
    part = line.split(',')
    frame.append(part[0])
    time.append('[' + part[3] + ':' + part[4] + ':' + part[5] + ']')
    value.append(part[10])
    if part[11]:
        ref1.append(part[11])
    if part[12]:
        ref2.append(part[12])


# noinspection PyUnusedLocal
def roteParser(line, frame, time, value, ref1, ref2, ref3, extra):
    """
    R-OTE sample line ---
        1.00,[17:25:39.3415],2737.8,3897.32,675.3,892.12
    """
    part = line.split(',')
    frame.append(part[0])
    time.append(part[1])
    value.append(part[2])
    if len(part) >= 4:
        if part[3]:
            ref1.append(part[3])
    if len(part) >= 5:
        if part[4]:
            ref2.append(part[4])
    if len(part) >= 6:
        if part[5]:
            ref3.append(part[5])


# noinspection PyUnusedLocal
def pymovieParser(line, frame, time, value, ref1, ref2, ref3, extra):
    """
    R-OTE sample line ---
        1.00,[17:25:39.3415],2737.8,3897.32,675.3,892.12
    """
    global pymovieDataColumns

    part = line.split(',')
    frame.append(part[0])
    time.append(part[1])
    dataColumnIndex = 0
    partNum = pymovieDataColumns[dataColumnIndex]
    value.append(part[partNum])
    dataColumnIndex += 1

    if len(part) >= 4 and pymovieSignalColumnCount >= 2:
        partNum = pymovieDataColumns[dataColumnIndex]
        if part[partNum]:
            ref1.append(part[partNum])
        dataColumnIndex += 1
    if len(part) >= 5 and pymovieSignalColumnCount >= 3:
        partNum = pymovieDataColumns[dataColumnIndex]
        if part[partNum]:
            ref2.append(part[partNum])
        dataColumnIndex += 1
    if len(part) >= 6 and pymovieSignalColumnCount >= 4:
        partNum = pymovieDataColumns[dataColumnIndex]
        if part[partNum]:
            ref3.append(part[partNum])
        dataColumnIndex += 1
    if pymovieSignalColumnCount > 4:
        for i in range(6, pymovieSignalColumnCount + 2):
            partNum = pymovieDataColumns[dataColumnIndex]
            if len(part) > i and part[partNum]:
                extra[i-6].append(part[partNum])
            dataColumnIndex += 1


# noinspection PyUnusedLocal
def rawParser(line, frame, time, value, secondary, ref2, ref3, extra):
    part = part = line.split(',')
    frame.append(part[0])
    time.append(part[1])
    value.append(part[2])


def readAs(file):
    global tangraNeedsBackgroundSubtraction
    global pymovieSignalColumnCount, pymovieDataColumns, pymovieColumnNamePrefix

    kind = getFileKind(file)
    
    fileobject = open(file)
    frame = []
    time = []
    value = []
    ref1 = []
    ref2 = []
    ref3 = []
    extra = []
    headers = []
    aperture_names = []
    
    if kind == 'Tangra':
        colHeaderKey = 'FrameNo'
        parser = tangraParser
    elif kind == 'R-OTE':  # PYOTE uses same colHeaderKey
        colHeaderKey = 'FrameNum'
        parser = roteParser
    elif kind == 'PyMovie':
        colHeaderKey = 'FrameNum'
        pymovieSignalColumnCount = 0
        parser = pymovieParser
    elif kind == 'Limovie':
        colHeaderKey = 'No.'
        parser = limovieParser
    elif kind == 'raw':
        colHeaderKey = 'RAW'
        parser = rawParser
    else:
        return False, 'invalid file "kind"', frame, time, value, ref1, ref2, ref3, extra, aperture_names, headers

    with fileobject:
        while True:
            line = fileobject.readline()
            if line:
                if colHeaderKey in line:

                    if kind == 'Tangra':
                        if line.find('SignalMinusBackground') > 0:
                            tangraNeedsBackgroundSubtraction = False
                        else:
                            tangraNeedsBackgroundSubtraction = True

                    if kind == 'PyMovie':
                        # We need to count the number of times 'signal" starts
                        # a column header
                        line = line.rstrip()  # Get rid of possible trailing new line \n
                        parts = line.split(',')
                        pymovieDataColumns = []
                        columnIndex = 0
                        for part in parts:
                            if part.startswith(pymovieColumnNamePrefix):
                                pymovieSignalColumnCount += 1
                                aperture_names.append(part.split('-')[1])
                                pymovieDataColumns.append(columnIndex)
                            columnIndex += 1
                            # If there are more than 4 columns of 'signals', we need to setup
                            # extra to hold those columns
                        for i in range(5, pymovieSignalColumnCount+1):
                            extra.append([])

                    while True:
                        line = fileobject.readline()
                        if line:
                            # noinspection PyBroadException
                            try:
                                parser(line, frame, time, value, ref1, ref2, ref3, extra)
                            except Exception as e:
                                return False, str(e), frame, time, value, \
                                       ref1, ref2, ref3, extra, aperture_names, headers
                        else:
                            return True, kind, frame, time, value, ref1, ref2, ref3, extra, aperture_names, headers
                headers.append(line[:-1])
            else:
                return (False, colHeaderKey + ' not found as first column header', 
                        [], [], [], [], [], [], [], aperture_names, headers)



import pyqtgraph as pg
import pyqtgraph.examples
from PyQt5 import QtGui
import cv2
import pyoteapp.SER
import glob
import astropy.io.fits as pyfits  # Used for reading/writing FITS files
from Adv2.Adv2File import Adv2reader
from Adv2.AdvError import AdvLibException


# def readAavFile(frame_to_read=0, full_file_path=None):
#     try:
#         rdr = Adv2reader(full_file_path)
#     except AdvLibException as adverr:
#         return repr(adverr), None
#     except IOError as ioerr:
#         return repr(ioerr), None
#
#     rdr.closeFile()
#     return 'ok', None


def readAavFile(frame_to_read=0, full_file_path=None):
    try:
        rdr = Adv2reader(full_file_path)
    except AdvLibException as adverr:
        return {"success": False, "image": None, "errmsg": repr(adverr)}
    except IOError as ioerr:
        return {"success": False, "image": None, "errmsg": repr(ioerr)}

    err, image, _, _ = rdr.getMainImageAndStatusData(frameNumber=frame_to_read)
    rdr.closeFile()
    if not err:
        return {"success": True, "image": image, "errmsg": ""}
    else:
        return {"success": False, "image": None, "errmsg": err}


def readAviFile(frame_to_read=0, full_file_path=None):
    cap = None
    fourcc = ''
    fps = None
    frame_count = None

    if full_file_path:

        cap = cv2.VideoCapture(full_file_path, cv2.CAP_FFMPEG)
        if not cap.isOpened():
            errmsg = f'{full_file_path} could not be opened!'
            return {"success": False, "image": None, "errmsg": errmsg,
                    "fourcc": '', "fps": 0.0, "num_frames": 0}
        else:
            # Let's get the FOURCC code
            fourcc = int(cap.get(cv2.CAP_PROP_FOURCC))
            fourcc_str = f'{fourcc & 0xff:c}{fourcc >> 8 & 0xff:c}{fourcc >> 16 & 0xff:c}{fourcc >> 24 & 0xff:c}'
            # print(f'FOURCC codec ID: {fourcc_str}')

            fps = cap.get(cv2.CAP_PROP_FPS)
            # print(f'frames per second: {fps:0.6f}')

            frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            # print(f'There are {frame_count} frames in the file.')

    # Read the specified frame
    success, image, errmsg = readAviFrame(frame_to_show=frame_to_read, cap=cap, fourcc=fourcc_str)

    if cap is not None:
        cap.release()

    return {"success": success, "image": image, "errmsg": errmsg,
            "fourcc": fourcc_str, "fps": fps, "num_frames": frame_count}


def readAviFrame(frame_to_show, cap, fourcc):
    image = None

    try:
        if fourcc == 'dvsd':
            success, frame = getDvsdAviFrame(frame_to_show, cap)
            if len(frame.shape) == 3:
                image = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        else:
            cap.set(cv2.CAP_PROP_POS_FRAMES, frame_to_show)
            status, frame = cap.read()
            image = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

        return True, image, 'ok'

    except Exception as e1:
        return False, None, f'Problem reading avi file: {e1}'


def getDvsdAviFrame(fr_num, cap):
    # This routine is used for 'dvsd' avi files because they cannot be positioned directly to a specific frame

    success = False
    frame = None

    next_frame = cap.get(cv2.CAP_PROP_POS_FRAMES)

    if fr_num == next_frame:
        success, frame = cap.read()
        return success, frame

    if fr_num > next_frame:
        frames_to_read = fr_num - next_frame + 1
        while frames_to_read > 0:
            frames_to_read -= 1
            success, frame = cap.read()
        return success, frame

    return False, None


def readSerFile(frame_to_read=0, full_file_path=None):

    if full_file_path:

        ser_meta_data, ser_timestamps = pyoteapp.SER.getMetaData(full_file_path)

        # showSerMetaData()
        frame_count = ser_meta_data['FrameCount']
        print(f'There are {frame_count} frames in the SER file.')
        bytes_per_pixel = ser_meta_data['BytesPerPixel']
        print(f'Image data is encoded in {bytes_per_pixel} bytes per pixel')

        try:
            bytes_per_pixel = ser_meta_data['BytesPerPixel']
            image_width = ser_meta_data['ImageWidth']
            image_height = ser_meta_data['ImageHeight']
            little_endian = ser_meta_data['LittleEndian']
            with open(full_file_path, 'rb') as ser_file_handle:
                image = pyoteapp.SER.getSerImage(
                    ser_file_handle, frame_to_read,
                    bytes_per_pixel, image_width, image_height, little_endian
                )
            raw_ser_timestamp = ser_timestamps[frame_to_read]
            parts = raw_ser_timestamp.split('T')
            time_stamp = f'{parts[0]} @ {parts[1]}'
            print(f'Timestamp found: {time_stamp}')
        except Exception as e:
            return {'success': False, 'errmsg': f"{e}", 'image': None, 'timestamp': ''}

        return {'success': True, 'errmsg': '', 'image': image, 'timestamp': time_stamp}


def readFitsFile(frame_to_read=0, full_file_path=None):
    fits_filenames = sorted(glob.glob(full_file_path + '/*.fits'))
    num_frames = len(fits_filenames)
    # print(f'Number of FITS frames: {num_frames}')
    errmsg = ''
    success = False
    time_stamp = ''

    try:
        hdr = pyfits.getheader(fits_filenames[frame_to_read], 0)

        image = pyfits.getdata(fits_filenames[frame_to_read], 0)

        if 'DATE-OBS' in hdr.keys():
            date_time = hdr['DATE-OBS']
            # The form of DATE-OBS is '2018-08-21T05:21:02.4561235' so we can simply 'split' at the T
            parts = date_time.split('T')
            time_stamp = f'{parts[0]} @ {parts[1]}'

        success = True

    except Exception as e3:
        errmsg = f'While reading image data from FITS file: {e3}'
        image = None

    return {'success': success, 'image': image, 'errmsg': errmsg, 'timestamp': time_stamp, 'num_frames': num_frames}


def excercise():
    _ = QtGui.QApplication([])

    win = QtGui.QMainWindow()
    win.resize(800, 800)
    imv = pg.ImageView()
    imv.show()

    pyqtgraph.examples.run()


if __name__ == "__main__":
    excercise()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'helpDialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(991, 434)
        Dialog.setSizeGripEnabled(True)
        self.gridLayout = QtWidgets.QGridLayout(Dialog)
        self.gridLayout.setObjectName("gridLayout")
        self.textEdit = QtWidgets.QTextEdit(Dialog)
        self.textEdit.setReadOnly(True)
        self.textEdit.setObjectName("textEdit")
        self.gridLayout.addWidget(self.textEdit, 0, 0, 1, 1)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Help"))
        self.textEdit.setToolTip(_translate("Dialog", "<html><head/><body><p>This dialog box is non-modal, so there is no need to close it in order to continue using the main GUI.</p><p><br/></p><p>In fact, best practice is to set your preferred size and position and DO NOT close the dialog.  When you click on the main GUI, this dialog box will appear behind the main GUI, but a new right-click click will pop it back up at the previous size and position.</p></body></html>"))
        self.textEdit.setHtml(_translate(
            "Dialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
            "<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
            "p, li { white-space: pre-wrap; }\n"
            "</style></head><body style=\" font-family:\'.SF NS Text\'; font-size:13pt; font-weight:400; font-style:normal;\">\n"
            "<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p></body></html>"))




"""
Created on Thu May 25 12:26:10 2017

@author: Bob Anderson
"""

import numpy as np
from PyQt5.QtWidgets import QDialog


def convertTimeStringToTime(timeStr):
    """
    We assume the timeStr has the format  [hh:mm:ss.ssss]
    Returns -1.0 if conversion fails, otherwise time as a float
    """
    # noinspection PyBroadException
    try:
        if timeStr[0] != '[':
            return -1.0
        if timeStr[-1] != ']':
            return -1.0
        clippedStr = timeStr[1:-1]  # remove [] from input
        parts = clippedStr.split(':')
        hourSecs = float(parts[0]) * 60 * 60
        minSecs = float(parts[1]) * 60
        secs = float(parts[2])
        return hourSecs + minSecs + secs
    except Exception:
        return -1.0
    

def convertTimeToTimeString(time):
    """
    Convert a time (float) to string --- [hh:mm:ss.ssss]
    """
    hours = int(time // (60 * 60))
    if hours > 23 or hours < 0:
        return ""
    time -= hours * 60 * 60
    minutes = int(time // 60)
    time -= minutes * 60
    seconds = time
    return '[{:02d}:{:02d}:{:07.4f}]'.format(hours, minutes, seconds)


def improveTimeStep(outliers, deltaTime):
    validIntervals = [deltaTime[i] for i in range(len(deltaTime)) if i not in outliers]
    # Under some circumstances, OCR can be so bad that no valid intervals can
    # be found.  In that case, we return a timeStep of 0.0 and let the
    # program respond to this invalid value at a higher level.
    if validIntervals:
        # Changed in 5.0.1
        # betterTimeStep = np.mean(validIntervals)
        betterTimeStep = np.median(validIntervals)
    else:
        betterTimeStep = 0.0
    return betterTimeStep


def getTimeStepAndOutliers(timestamps, yValues, yStatus, VizieRdict=None):

    # yValue is a numpy array; timestamps and yStatus are lists

    time = [convertTimeStringToTime(item) for item in timestamps]
    deltaTime = np.diff(time)

    # Changed in 5.0.1
    # timeStep = np.median(deltaTime)
    timeStep = np.mean(deltaTime)
    cadence_high = timeStep * 1.3  # Had been 1.2
    cadence_low = timeStep * 0.7   # Had been 0.8

    dropped_frame_high = timeStep * 1.8
    dropped_frame_low = timeStep * 0.2

    droppedFrameIndices = [i for i, dt in enumerate(deltaTime)
                           if dt < dropped_frame_low or dt > dropped_frame_high]
    numEntries = len(timestamps)
    numOutliers = len(droppedFrameIndices)
    timestampErrorRate = numOutliers / numEntries
    
    improvedTimeStep = improveTimeStep(droppedFrameIndices, deltaTime)

    cadenceViolationIndices = [i for i, dt in enumerate(deltaTime)
                               if dt < cadence_low or dt > cadence_high]

    # Calculate number of dropped frames at each droppedFrameIndices point
    timingReport = []
    showDetails = True
    if showDetails:
        cumDroppedReadings = 0
        cumDuplicatedReading = 0
        cumTimestampErrors = 0
        for index in droppedFrameIndices:
            timeChange = deltaTime[index]
            droppedReadings = round(timeChange / improvedTimeStep) - 1
            if droppedReadings > 0:
                cumDroppedReadings += droppedReadings
                timingReport.append(f'At reading {index:05}: time delta to next reading = {timeChange:0.4f} : likely {droppedReadings} dropped readings')
            elif droppedReadings < 0:
                timingReport.append(f'At reading {index:05}: time delta to next reading = {timeChange:0.4f} : likely a timestamp error')
                cumTimestampErrors += 1
            else:
                cumDuplicatedReading += 1
                timingReport.append(f'At reading {index:05}: time delta to next reading = {timeChange:0.4f} : likely a duplicated reading')

        if cumDroppedReadings > 0 or cumDuplicatedReading > 0 or len(cadenceViolationIndices) > 0:
            timingReport.append('')
            timingReport.append(f'========== Total dropped readings..... {cumDroppedReadings}')
            timingReport.append(f'========== Num duplicated readings.... {cumDuplicatedReading}')
            timingReport.append(f'========== Number timestamp errors.... {cumTimestampErrors}')
            timingReport.append(f'========== Number of cadence errors... {len(cadenceViolationIndices)}')
            timingReport.append('')

    # Added for exporting lightcurve in archive format (only used during initial read of file)
    if VizieRdict is not None:
        newTimeStamps, newYvalues, newYstatus = insertDroppedReadings(timestamps, yValues, yStatus, improvedTimeStep)
        VizieRdict["timestamps"] = newTimeStamps
        VizieRdict["yValues"] = newYvalues
        VizieRdict["yStatus"] = newYstatus

    return improvedTimeStep, droppedFrameIndices, cadenceViolationIndices, timestampErrorRate, timingReport


# Status of points and associated dot colors ---
MISSING = 5   # (211, 211, 211) light gray
EVENT = 4     # (155, 150, 100) sick green
SELECTED = 3  # (255, 000, 000) red
BASELINE = 2  # (255, 150, 100) salmon
INCLUDED = 1  # (000, 032, 255) blue with touch of green
EXCLUDED = 0  # no dot

MISSING_READING_VALUE = -0.0

def invalidStep(tNow, tNext, timeStep):
    dropped_reading_high = timeStep * 1.8
    dropped_reading_low = timeStep * 0.2
    dt = tNext - tNow
    return dt < dropped_reading_low or dt > dropped_reading_high

def insertDroppedReadings(timestamps, yValues, yStatus, timeStep):
    # print(f'timestamps: {type(timestamps)}  yValues: {type(yValues)}  yStatus: {type(yStatus)}')

    tNow = convertTimeStringToTime(timestamps[0])
    timestampsExpanded = [timestamps[0]]
    yValuesExpanded = [yValues[0]]
    yStatusExpanded = [INCLUDED]
    k = 1

    while k < yValues.size:
        tNext = convertTimeStringToTime((timestamps[k]))
        stepOk = not invalidStep(tNow, tNext, timeStep)
        if stepOk:
            yValuesExpanded.append(yValues[k])
            timestampsExpanded.append(timestamps[k])
            yStatusExpanded.append(INCLUDED)
            tNow = tNext
        else:
            # Compute number of dropped readings
            timeChange = tNext - tNow
            numDroppedReadings: int = round(timeChange / timeStep) - 1

            if numDroppedReadings < 0:
                # This is likely a timestamp reading error. Ignoring it is the best we can do.
                numDroppedReadings = 0
            else:
                tNext = tNow + timeStep
                for i in range(numDroppedReadings):
                    tNext = tNow + timeStep
                    yValuesExpanded.append(MISSING_READING_VALUE)
                    timestampsExpanded.append(convertTimeToTimeString(tNext))
                    yStatusExpanded.append(MISSING)
                    tNow = tNext
                k -= 1
        k += 1

    return timestampsExpanded, np.array(yValuesExpanded), yStatusExpanded


def isFrameNumberInData(frameToTest, frame):
    first = float(frame[0])
    last = float(frame[-1])
    return first <= frameToTest <= last


savedF1 = savedHH1 = savedMM1 = savedSS1 = ''
savedF2 = savedHH2 = savedMM2 = savedSS2 = ''
savedNTSCButton = True
savedPALButton = False
savedCustomButton = False
savedCustomText = ''


def saveAllDialogEntries(dialog):
    global savedF1, savedHH1, savedMM1, savedSS1
    global savedF2, savedHH2, savedMM2, savedSS2
    global savedNTSCButton, savedPALButton, savedCustomButton
    global savedCustomText

    savedF1 = dialog.frameNum1.text()
    savedHH1 = dialog.hh1.text()
    savedMM1 = dialog.mm1.text()
    savedSS1 = dialog.ss1.text()

    savedF2 = dialog.frameNum2.text()
    savedHH2 = dialog.hh2.text()
    savedMM2 = dialog.mm2.text()
    savedSS2 = dialog.ss2.text()

    savedNTSCButton = dialog.radioButtonNTSC.isChecked()
    savedPALButton = dialog.radioButtonPAL.isChecked()
    savedCustomButton = dialog.radioButtonCustom.isChecked()

    savedCustomText = dialog.frameDeltaTime.text()


# noinspection PyBroadException,PyUnusedLocal
def manualTimeStampEntry(frame, dialog, flashFrames=None):
    time = []  # Output --- to be computed from timestamp data entered in dialog
    dataEntered = 'error in data entry'

    if flashFrames:
        dialog.frameNum1.setText(str(flashFrames[0]))
        if len(flashFrames) > 1:
            dialog.frameNum2.setText(str(flashFrames[1]))
    else:
        dialog.frameNum1.setText(savedF1)
        dialog.frameNum2.setText(savedF2)

    dialog.hh1.setText(savedHH1)
    dialog.mm1.setText(savedMM1)
    dialog.ss1.setText(savedSS1)

    dialog.hh2.setText(savedHH2)
    dialog.mm2.setText(savedMM2)
    dialog.ss2.setText(savedSS2)

    if savedNTSCButton:
        dialog.radioButtonNTSC.setChecked(savedNTSCButton)
    if savedPALButton:
        dialog.radioButtonPAL.setChecked(savedPALButton)
    if savedCustomButton:
        dialog.radioButtonCustom.setChecked(savedCustomButton)

    dialog.frameDeltaTime.setText(savedCustomText)

    result = dialog.exec_()

    saveAllDialogEntries(dialog)

    if result == QDialog.Accepted:
        nf = ef = frameNum1 = frameNum2 = frameDeltaTime = -1
        hh1 = mm1 = ss1 = hh2 = mm2 = ss2 = -1
        try:
            frameNum1 = float(dialog.frameNum1.text())
        except Exception:
            return '"' + dialog.frameNum1.text() + '" is invalid as frame number input', \
                   time, dataEntered, nf, ef
        try:
            hh1 = int(dialog.hh1.text())
        except Exception:
            return '"' + dialog.hh1.text() + '" is invalid as hh input', \
                   time, dataEntered, nf, ef
        try:
            mm1 = int(dialog.mm1.text())
        except Exception:
            return '"' + dialog.mm1.text() + '" is invalid as mm input', \
                   time, dataEntered, nf, ef
        try:
            ss1 = float(dialog.ss1.text())
        except Exception:
            return '"' + dialog.ss1.text() + '" is invalid as ss.ssss input', \
                   time, dataEntered, nf, ef

        frameNum2text = dialog.frameNum2.text()
        if frameNum2text != '':
            try:
                frameNum2 = float(dialog.frameNum2.text())
            except Exception:
                return '"' + dialog.frameNum2.text() + '" is invalid as frame number input', \
                       time, dataEntered, nf, ef
            try:
                hh2 = int(dialog.hh2.text())
            except Exception:
                return '"' + dialog.hh2.text() + '" is invalid as hh input', \
                       time, dataEntered, nf, ef
            try:
                mm2 = int(dialog.mm2.text())
            except Exception:
                return '"' + dialog.mm2.text() + '" is invalid as mm input', \
                       time, dataEntered, nf, ef
            try:
                ss2 = float(dialog.ss2.text())
            except Exception:
                return '"' + dialog.ss2.text() + '" is invalid as ss.ssss input', \
                       time, dataEntered, nf, ef
        else:
            return 'Both entries must be supplied', \
                   time, dataEntered, nf, ef

        if dialog.frameDeltaTime.text() and not dialog.radioButtonCustom.isChecked():
            return 'You have something entered in custom frame time edit box ' + \
                   'but have not clicked on the radio button to enable use of this value.\n\n' + \
                   'Please clarify your intentions.', \
                   time, dataEntered, nf, ef

        if dialog.radioButtonNTSC.isChecked():
            expectedFrameDeltaTime = 1.001 / 30.0
        elif dialog.radioButtonPAL.isChecked():
            expectedFrameDeltaTime = 1.000 / 25.0
        else:
            try:
                expectedFrameDeltaTime = float(eval(dialog.frameDeltaTime.text(), {}, {}))
            except Exception:
                return '"' + dialog.frameDeltaTime.text() + '" is invalid as timeDelta', \
                       time, dataEntered, nf, ef
            if not isinstance(expectedFrameDeltaTime, float):
                return '"' + dialog.frameDeltaTime.text() + '" is invalid as timeDelta --- not a float', \
                       time, dataEntered, nf, ef
            if not expectedFrameDeltaTime > 0.0:
                return '"' + dialog.frameDeltaTime.text() + '" is invalid as timeDelta --- not > 0', \
                       time, dataEntered, nf, ef

        # Validate data entries

        if frameNum2 <= frameNum1:
            return 'early frame num must be less than late frame num', \
                   time, dataEntered, nf, ef
        if frameNum1 < 0 or hh1 < 0 or mm1 < 0 or ss1 < 0:
            return 'Negative values of frame num are invalid.', \
                   time, dataEntered, nf, ef
        if frameNum2 < 0 or hh2 < 0 or mm2 < 0 or ss2 < 0:
            return 'Negative values of frame num are invalid.', \
                   time, dataEntered, nf, ef
        if not isFrameNumberInData(frameNum1, frame):
            return 'early frame num is not valid: could not be found in the file data', \
                   time, dataEntered, nf, ef
        if not isFrameNumberInData(frameNum2, frame):
            return 'late frame num is not valid: could not be found in the file data', \
                   time, dataEntered, nf, ef

        timeStr1 = 'Manual timestamp info: @ frame {:0.2f} [{:02d}:{:02d}:{:07.4f}]'.format(frameNum1, hh1, mm1, ss1)
        timeStr2 = ' --- @ frame {:0.2f} [{:02d}:{:02d}:{:07.4f}]'.format(frameNum2, hh2, mm2, ss2)
        dataEntered = timeStr1 + timeStr2

        numFramesInSpan = frameNum2 - frameNum1
        t1 = hh1 * 3600 + mm1 * 60 + ss1
        t2 = hh2 * 3600 + mm2 * 60 + ss2
        # Handle midnight crossing
        if t2 < t1:
            t2 += 3600.0 * 24.0
        expectedFramesInSpan = (t2 - t1) / expectedFrameDeltaTime

        # Time to compute timestamps.
        calculatedFrameDeltaTime = (t2 - t1) / (frameNum2 - frameNum1)
        time = timestampsFromFrameDelta(t1, frameNum1, calculatedFrameDeltaTime, frame)

        return 'ok', time, dataEntered, numFramesInSpan, expectedFramesInSpan
    else:
        return 'cancelled', time, dataEntered, -1, -1


def timestampsFromFrameDelta(t1, frameNum1, frameDeltaTime, frame):
    time = []
    secondsIn24Hours = 3600 * 24.0
    for item in frame:
        frameNum = float(item)
        timeOfFrame = t1 + (frameNum - frameNum1) * frameDeltaTime
        if timeOfFrame >= secondsIn24Hours:
            timeOfFrame -= secondsIn24Hours
        elif timeOfFrame < 0:
            timeOfFrame += secondsIn24Hours
        time.append(convertTimeToTimeString(timeOfFrame))

    return time



# def upgradePyote(pyoteversion):
#
#     import subprocess
#
#     try:
#         import pipenv
#         resp = subprocess.run(['python3', '-m', 'pipenv', 'install', pyoteversion],
#                               stderr=subprocess.PIPE, stdout=subprocess.PIPE)
#     except Exception:
#         resp = subprocess.run(['python', '-m', 'pip', 'install', '--user', '--upgrade', pyoteversion],
#                               stderr=subprocess.PIPE, stdout=subprocess.PIPE)
#
#     return resp.stdout.decode("utf-8").split('\n')


# The following fuction was added 23 July 2022 when (apparently) the PyPI JSON API broke.
# This uses a reliable, supported technique to get info (we are only interested in Version: )
# about a package on PyPI
def getLatestPackageVersion(package_name: str) -> str:
    import subprocess
    try:
        response = subprocess.run(['python', '-m', 'pip', 'install', f"{package_name}==0.0.0"],
                                  stderr=subprocess.PIPE, stdout=subprocess.PIPE)
        errorResponse = response.stderr.decode("utf-8")
        versions = errorResponse.split('versions: ')
    except FileNotFoundError as e:
        print(f'{e}')
        versions = [0]
    if len(versions) == 1:  # Because the split above failed
        print('python not used to start pyote or no internet connection')
        try:
            response = subprocess.run(['python3', '-m', 'pip', 'install', f"{package_name}==0.0.0"],
                                      stderr=subprocess.PIPE, stdout=subprocess.PIPE)
            errorResponse = response.stderr.decode("utf-8")
            versions = errorResponse.split('versions: ')
        except FileNotFoundError as e:
            print(f'{e}')
            versions = [0]
        if len(versions) == 1:
            print('python3 not used to start pyote or no internet connection')
            try:
                response = subprocess.run(['py', '-m', 'pip', 'install', f"{package_name}==0.0.0"],
                                          stderr=subprocess.PIPE, stdout=subprocess.PIPE)
                errorResponse = response.stderr.decode("utf-8")
                versions = errorResponse.split('versions: ')
            except FileNotFoundError as e:
                print(f'{e}')
                versions = [0]
            if len(versions) == 1:
                print('py not used to start pyote or no internet connection')
                # Failed to make Internet connection
                return 'Failed to connect to PyPI - Internet connection problem?'
    versions = versions[1].split(')')[0]  # Remove everything at and after ')'
    latestVersion = versions.split(',')[-1].strip()

    return latestVersion



from numpy import zeros, complex128
from numba import jit, prange
import numpy as np
from scipy.fftpack import fft2
from numpy import exp, pi
from scipy.special import fresnel
from scipy import interpolate


@jit(nopython=True)
def tresterModifiedAperture(aperture, N):
    modified_aperture = zeros((N, N), dtype=complex128)
    c1 = 1.0j * pi / N
    for i in prange(N):
        for k in prange(N):
            c2 = i * i + k * k
            modified_aperture[i, k] = aperture[i, k] * exp(c1 * c2)
    return modified_aperture


def fresnelLength(wavelength_nm, Z_AU):
    wavelength = wavelength_nm * 1e-9 * 1e-3
    Z = Z_AU * 149.6e6
    return np.sqrt(wavelength * Z / 2)


def convert_fresnel_distance_to_pixels(fresnel_distance, N):
    return round(fresnel_distance * np.sqrt(N / 2))


def convert_km_distance_to_pixels(km_distance, N, fresnel_length_km):
    return convert_fresnel_distance_to_pixels(km_distance / fresnel_length_km, N)


def diffraction_u(u: float, type_D_or_R: str) -> float:
    """
    Calculate normalized knife edge diffraction curve for a single wavelength using fresnel unit scale.

    When type_D_or_R == 'R' then a 'reappearance' curve
    results: it is 'dark' to the left of zero, and 'light' to the right of zero.

    When type_D_or_R == 'D' then a 'disappearance' curve
    results: it is 'light' to the left of zero, and 'dark' to the right of zero.

    :param u: fresnel unit
    :param type_D_or_R: either 'D' (disappearance) or 'R' (reappearance)
    :return: Normalized, single wavelength, diffraction intensity as a function of fresnel units
    """
    assert type_D_or_R in ['D', 'R'], 'Invalid type code given to diffraction_u'

    if type_D_or_R == 'D':
        u = -u
    ssa, csa = fresnel(u)
    return 0.5 * ((0.5 + csa)**2 + (0.5 + ssa)**2)


def analyticDiffraction(u_value, side='R'):
    return diffraction_u(u_value, side)


# We use Numba to greatly speedup the calculation of the modified aperture.
@jit(nopython=True)
def basicCalcField(N, fresnel_length_km, diam_km):
    # None of the following values are used in the computation of the modified aperture field
    # in the plane of the asteroid, but are needed to interpret the result in useful
    # physical parameters of field-of-view (in kilometers) and fresnel lengths, so we compute
    # these values and provide them in the returned values

    elz = 2 * fresnel_length_km * fresnel_length_km
    L_km = np.sqrt(elz * N)  # km  (We're viewing an L_km x L_km image) == fresnel_length_km * sqrt(2*N) == FOV
    # For N = 2048 the FOV = 64 fresnel lengths on a side
    x = np.linspace(-L_km / 2, L_km / 2, N)
    x_fl = x / fresnel_length_km  # x_fl is x expressed in fresnel lengths

    field = np.zeros((N, N), dtype=np.complex128)  # Ensure the field is complex

    amplitude = 1.0 + 0.0j

    r = diam_km / 2.0
    r2 = r * r

    c1 = 1j * pi / N

    for i in range(N):
        for j in range(N):
            c2 = i * i + j * j
            in_target = (x[j] * x[j] + x[i] * x[i]) < r2
            in_target = in_target or (x[j] * x[j] + x[i] * x[i]) < r2
            if not in_target:
                field[i, j] = amplitude * np.exp(c1 * c2)

    return field, fresnel_length_km, L_km, x_fl


def generalizedDiffraction(asteroid_diam_km, asteroid_distance_AU,
                           graze_offset_km=0, wavelength1=400, wavelength2=600,
                           skip_central_flash=False):
    # A family of central spot lightcurves are calculated when rho <= 32, otherwise the
    # analytic diffraction equation is used to produce the curve family that is
    # subsequently integrated.

    # Single frequency lightcurves can be generated by setting wavelength1 = wavelength2 at
    # the desired wavelength.

    # Grazing observation lightcurves can be generated with any amout of offset, including
    # outside the geometrical shadow.

    y_avg = []
    x_avg = []

    rows_y = []
    rows_x = []
    rows_u = []

    wavelength_steps = 10

    N = 2048  # This gives a FOV of 64 fresnel lengths

    max_n_off = N // 2 - 1

    wavelength1 = int(wavelength1)
    wavelength2 = int(wavelength2)

    rho_wavelength = (wavelength1 + wavelength2) // 2

    if wavelength1 == wavelength2:
        wavelengths = [wavelength1]
        single_wavelength = True
    else:
        wavelengths = [i for i in range(wavelength1, wavelength2 + 1, wavelength_steps)]
        single_wavelength = len(wavelengths) == 1

    # asteroid_radius_km = asteroid_diam_km / 2

    # Compose title
    title = f'graze-{graze_offset_km:0.2f}km_astdiam-{asteroid_diam_km:0.2f}km_astdist-{asteroid_distance_AU}AU'
    if not single_wavelength:
        title += f'_{wavelength1}-to-{wavelength2}nm'
    else:
        title += f'_{wavelength1}nm'

    asteroid_radius_km = asteroid_diam_km / 2

    rho = asteroid_diam_km / fresnelLength(wavelength_nm=rho_wavelength, Z_AU=asteroid_distance_AU)

    if rho <= 32 and not skip_central_flash:
        for wavelength in wavelengths:
            fresnel_length_km = fresnelLength(wavelength_nm=wavelength, Z_AU=asteroid_distance_AU)
            if graze_offset_km == 0.0:
                n_off = 0
            else:
                n_off = convert_km_distance_to_pixels(graze_offset_km, N, fresnel_length_km)
                if n_off > max_n_off:
                    n_off = max_n_off
            field, fresnel_length_km, L_km, x_fl = basicCalcField(N, fresnel_length_km, diam_km=asteroid_diam_km)
            transform = (fft2(field)) / N
            image = abs(transform) ** 2
            row = image[N // 2 + n_off, :]
            x_km = x_fl * fresnel_length_km

            rows_y.append(row)
            rows_x.append(x_km)
            rows_u.append(x_fl)
    else:
        plot_span = 3
        x_km = np.linspace(-plot_span * asteroid_radius_km, plot_span * asteroid_radius_km, N)
        z_km = np.sqrt(graze_offset_km ** 2 + x_km ** 2)

        for wavelength in wavelengths:
            fresnel_length_km = fresnelLength(wavelength_nm=wavelength, Z_AU=asteroid_distance_AU)

            x_fl = x_km / fresnel_length_km

            z_fl = z_km / fresnel_length_km

            y = []
            asteroid_radius_u = asteroid_radius_km / fresnel_length_km
            for u_value in z_fl:
                y.append(analyticDiffraction(u_value - asteroid_radius_u, 'R'))

            rows_y.append(y)
            rows_x.append(x_km)
            rows_u.append(x_fl)

    k = len(rows_x)

    # Generate an interpolation function f for each wavelength sample (not needed for analyticDiffraction,
    # but it doesn't slow the calculation in a perceptible way)
    interpolation_function = []
    for i in range(k):
        interpolation_function.append(interpolate.interp1d(rows_x[i], rows_y[i], kind='quadratic', bounds_error=True))

    # Perform integration (using interpolation functions generated above)
    for x in rows_x[0]:
        x_avg.append(x)
        cum_sum = 0.0
        for i in range(k):
            # interpolate so that we can sum contributions at identical positions
            cum_sum += interpolation_function[i](x)
        y_avg.append(cum_sum / k)

    return np.array(x_avg), np.array(y_avg), rho, rho_wavelength, wavelengths[0], wavelengths[-1], title


# ############################## Utility functions ###############################
def asteroid_mas(diameter_km, distance_AU):
    milli_arcseconds = 1000 * 206265 * diameter_km / (1.496e8 * distance_AU)
    return milli_arcseconds


def distance_AU_from_parallax_arcsec(parallax_arcsec):
    return 8.7882 / parallax_arcsec


def projected_star_diameter_km(star_mas, asteroid_distance_AU):
    diam = 1.496e8 * asteroid_distance_AU * star_mas / (1000 * 206265)
    return diam


def star_diameter_fresnel(wavelength_nm, star_mas, asteroid_distance_AU):
    fresnel_length_km = fresnelLength(wavelength_nm=wavelength_nm, Z_AU=asteroid_distance_AU)
    star_diam_km = projected_star_diameter_km(star_mas=star_mas, asteroid_distance_AU=asteroid_distance_AU)
    return star_diam_km / fresnel_length_km


# Here we codify the decision process as to which lightcurve model to use
def decide_model_to_use(asteroid_diameter_km=5,
                        asteroid_distance_AU=2,
                        star_diameter_mas=0.0,
                        shadow_speed=5,
                        frame_time=0.0334):

    print(f'asteroid_diameter_km: {asteroid_diameter_km:0.3f}')
    star_diameter_km = projected_star_diameter_km(star_diameter_mas, asteroid_distance_AU)
    print(f'star_diameter_km: {star_diameter_km:0.3f}')

    fresnel_length = fresnelLength(wavelength_nm=500, Z_AU=asteroid_distance_AU)
    print(f'fresnel_length: {fresnel_length:0.3f}')

    fresnel_time_sec = fresnel_length / shadow_speed
    print(f'\n')
    print(f'fresnel_time_sec: {fresnel_time_sec:0.3f}')
    print(f'frame_time: {frame_time:0.3f} sec')
    print(f'\n')

    if star_diameter_km > fresnel_length:  # This integrates away diffraction wiggles
        if asteroid_diameter_km >= 10 * star_diameter_km:
            print('Use edge-on-disk model because asteroid diameter is at least 10 times the star diameter')
        else:
            print('Use disk-on-disk model because asteroid diameter is less than 10 time the star diameter')
    elif frame_time < fresnel_time_sec:  # We are sampling fast enough to see diffraction wiggles
        print('Use diffraction model')
    else:
        print('Use square-wave model because diffraction effects are integrated away')



import numpy as np
from numpy import random, convolve
from numba import prange, njit, int64, float64
# import time
import matplotlib
# import matplotlib.pyplot as plt
# from pyoteapp.autocorrtools import *
matplotlib.use('Qt5Agg')


@njit(float64[:](int64, float64), cache=True)
def noise_gen_jit(num_points: int, noise_sigma: float) -> np.ndarray:
    out = np.empty(num_points)
    for i in range(num_points):
        out[i] = random.normal(0.0, noise_sigma)
    return out


@njit(cache=True)
def simple_convolve(x, c):
    nx = len(x)
    nc = len(c)
    nout = nx - nc + 1
    out = np.zeros(nout)
    for i in range(nout):
        for j in range(nc):
            out[i] += x[i + j] * c[j]
    return out


@njit(cache=True)
def simple_convolve2(x, c):
    nx = len(x)
    nc = len(c)
    nout = nx - nc + 1
    out = np.zeros(nout)
    i = 0
    while i < nout:
        j = 0
        while j < nc:
            out[i] += x[i + j] * c[j]
            j += 1
        i += 1
    return out


def noise_gen_numpy(num_points: int, noise_sigma: float) -> np.ndarray:
    return random.normal(size=num_points, scale=noise_sigma)


# @jit(float64[:](int64, float64, float64[:]))
def simulated_observation(observation_size: int, noise_sigma: float, corr_array: np.ndarray) -> np.ndarray:
    """Returns a numpy array of floats representing an observation that contains only correlated noise.

    observation_size: number of points in the simulated observation
    noise_sigma:      standard deviation of observation noise
    corr_array:      correlation array needed to produce the correlated noise
    """
    return convolve(noise_gen_jit(observation_size, noise_sigma), corr_array, 'same')
    # noise_array = noise_gen_jit(observation_size+len(corr_array), noise_sigma)
    # out = np.empty(observation_size)
    # for i in prange(observation_size):
    #     out[i] = np.dot(corr_array, noise_array[i:i+len(corr_array)])
    #
    # return out


# noinspection PyPep8Naming
@njit(cache=True)
def max_drop_in_simulated_observation(
        event_duration: int,
        observation_size: int,
        noise_sigma: float,
        corr_array: np.ndarray) -> float:
    """Returns the maximum drop from a simulated observation.

    An observation consisting only of correlated gaussian noise is
    created and exhaustively searched for an 'event' of a specified size.
    The biggest 'drop' (B - A) found is returned.  This drop will always
    be positive (>= 0.0)

    event_duration:   size/duration of event being simulated
    observation_size: number of points in the simulated observation
    noise_sigma:      standard deviation of observation noise
    corr_array:       correlation array needed to produce the correlated noise
    """

    # obs = simulated_observation(observation_size, noise_sigma, corr_array)
    obs = noise_gen_jit(observation_size, noise_sigma)
    obs = simple_convolve(obs, corr_array)

    obs_size = len(obs)

    numA = event_duration
    numB = obs_size - numA

    # Start with the trial 'event' position at the extreme left of the observation
    cumA = np.sum(obs[0:numA])
    cumB = np.sum(obs[numA:])

    A = cumA / numA
    B = cumB / numB

    drop = B - A

    # Use negative values returned by this routine as an indicator that no positive drop could be found
    # best_drop_so_far = drop if drop >= 0.0 else 0.0
    best_drop_so_far = -1.0

    # Calculate new values for B and A iteratively and conditionally update best_drop_so_far
    for i in range(1, numB):
        cumA -= obs[i]
        cumB += obs[i]

        cumA += obs[i + event_duration]
        cumB -= obs[i + event_duration]

        A = cumA / numA
        B = cumB / numB

        drop = B - A
        if drop > best_drop_so_far:
            best_drop_so_far = drop

    return best_drop_so_far


@njit(parallel=True)
def compute_drops(event_duration: int,
                  observation_size: int,
                  noise_sigma: float,
                  corr_array: np.ndarray,
                  num_trials: int) -> np.ndarray:
    """Returns the maximum drops from num_trials simulated observations.

    event_duration:   size/duration of event being simulated
    observation_size: number of points in the simulated observation
    noise_sigma:      standard deviation of observation noise
    corr_array:       correlation array needed to produce the correlated noise
    num_trials:       number of simulated observations
    """

    drops: np.ndarray = np.empty(num_trials)

    for i in prange(num_trials):
        drops[i] = max_drop_in_simulated_observation(
            event_duration=event_duration, observation_size=observation_size,
            noise_sigma=noise_sigma, corr_array=corr_array)

    return drops


# def excercise():
#     noise_gen_jit(1, 1.0)
#     noise_gen_jit(1, 1.0)
#
#     start_time = time.time()
#     _ = noise_gen_jit(4_000, 1.0)
#     stop_time = time.time()
#     print(f'noise_gen_jit time: {stop_time - start_time:.6f} seconds')
#
#     start_time = time.time()
#     _ = noise_gen_numpy(4_000, 1.0)
#     stop_time = time.time()
#     print(f'noise_gen_numpy time: {stop_time - start_time:.6f} seconds')
#
#     # Warm-up calls to allow Numba compiler to do its work
#     print(f'\nDoing Numba warm-up calls...')
#     simulated_observation(100, 1.0, np.ones(10))
#     simulated_observation(100, 1.0, np.ones(10))
#
#     max_drop_in_simulated_observation(10, 100, 1.0, np.ones(10))
#     max_drop_in_simulated_observation(10, 100, 1.0, np.ones(10))
#
#     compute_drops(10, 100, 1.0, np.ones(10), 1)
#     compute_drops(10, 100, 1.0, np.ones(10), 1)
#     print(f'...finished warm-up calls.\n')
#     # End warm-up calls
#
#     dur = 100
#     n_obs = 4000
#     sigma = 2.0
#     # corr = np.ones(10)
#     corr = np.array([0.11653067, 0.29474461, 0.40390159, 0.85814318])
#     n_trials = 10000
#
#     start_time = time.time()
#     drop_array = compute_drops(event_duration=dur, observation_size=n_obs,
#                                noise_sigma=sigma, corr_array=corr, num_trials=n_trials)
#     stop_time = time.time()
#     print(f'{n_trials} drops computed in {stop_time - start_time:.6f} seconds')
#
#     # %%
#     plt.plot(np.sort(drop_array))
#     plt.show()
#
#     # %%
#     hist_values, bin_edges = np.histogram(drop_array, 50)
#     # print(hist_values)
#     # print(bin_edges)
#     plt.hist(drop_array, bins=50)
#     max_drop = np.max(drop_array)
#     plt.vlines(max_drop, 0, max(hist_values), colors='r')
#     plt.show()
#
#     noise_gen = CorrelatedNoiseGenerator(acfcoeffs=[1.0, 0.5, 0.3, 0.1])
#     test_noise = noise_gen.corr_normal(num_values=100000)
#     print(f'sigma: {np.std(test_noise)}')
#     print(autocorr(test_noise, lastlag=6))
#     print(noise_gen.final_coeffs)
#

# if __name__ == "__main__":
#     excercise()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat May 27 20:11:13 2017

@author: bob
"""

import numpy as np
from scipy.signal import savgol_filter as savgol


def laggedCoef(x, lag):
    if lag == 0:
        return 1.0
    corMatrix = np.corrcoef(x[lag:], x[:-lag])
    return corMatrix[0, 1]


def savgolTrendLine(y, window=101, degree=3):

    if window > len(y):
        window = len(y)

    # savgol requires an odd number for the window --- enforce that here
    if window % 2 == 0:
        window -= 1
        
    stage1trend = savgol(np.array(y), window, degree)
    stage2trend = savgol(stage1trend, window, degree)

    return stage2trend  # This is a numpy.ndarray


def polyTrendLine(x, y, degree=3):
    poly = np.polyfit(x, y, degree)
    return np.polyval(poly, x)


def getCorCoefs(x, y):
    combo = list(zip(x, y))
    combo.sort()
    yvalsConcat = np.array([item[1] for item in combo])
    try:
        trend = savgolTrendLine(yvalsConcat, window=301, degree=1)
        residuals = yvalsConcat - trend
    except ValueError:
        return np.array([1.0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]), len(x), 0.0

    # We only compute correlation coefficients if there at least 14 data points
    if len(residuals) < 14:
        ans = [1.0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    else:
        ans = []
        for lag in range(11):
            ans.append(laggedCoef(residuals, lag))

    return np.array(ans), len(x), np.std(residuals)
